'''Distance matrix between two sets of time series.

The original code was written by D. Bueso and was taken from [2]_.

Original reference:
- Bueso, D., Piles, M. & Camps-Valls, G. Nonlinear PCA for Spatio-Temporal
Analysis of Earth Observation Data. IEEE Trans. Geosci. Remote Sensing 1–12
(2020) doi:10.1109/TGRS.2020.2969813.

References
----------
.. [1] https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=8989964&casa_token=3zKG0dtp-ewAAAAA:FM1CrVISSSqhWEAwPGpQqCgDYccfLG4N-67xNNDzUBQmMvtIOHuC7T6X-TVQgbDg3aDOpKBksg&tag=1
.. [2] https://github.com/DiegoBueso/ROCK-PCA

'''
import numpy as np


def distance_matrix(X1, X2, algorithm='original'):
    D = - 2 * (X1.conj().T @ X2)
    if algorithm == 'original':
        D = D + np.sum(X1 * X1, axis=0)[..., None]
        D = D + np.sum(X2 * X2, axis=0)
    # NOTE: Is this implementation really correct?
    # The distance matrix is not zero on the diagonal...
    # Isn't there a conjugate missing in the original formulation?
    else:
        D = D + np.sum(X1 * X1.conj(), axis=0)[..., None]
        D = D + np.sum(X2 * X2.conj(), axis=0)
    return D



from typing import Optional, Tuple, Union, List

import pandas as pd
import xarray as xr

from ..models._base_mca import _BaseMCA
from ..utils.tools import squeeze
from ._transformer import _MultiDataArrayTransformer

DataArray = xr.DataArray
DataArrayList = Union[DataArray, List[DataArray]]


class MCA(_BaseMCA):
    '''Maximum Covariance Analysis of two or multiple ``xr.DataArray``.

    Parameters
    ----------
    X, Y : DataArrayList
        Data for which to maximise covariance.
    axis : int
        Axis along which variance should be maximsed (the default is 0).
    n_modes : Optional[int]
        Number of modes to compute. Computing less modes can results in
        performance gains. If None, then the maximum number of modes is
        equivalent to ``min(n_samples, n_features)`` (the default is None).
    norm : bool
        Normalize each feature (e.g. grid cell) by its temporal standard
        deviation (the default is False).
    weights_X, weights_Y : Optional[DataArrayList] = None
        Weights applied to features of ``X`` and ``Y``. Must have the same
        dimensions as the original features which are the remaining axes
        not specified by ``axis`` parameter).


    Examples
    --------

    Import package and create data:

    >>> import xarray as xr
    >>> from xeofs.xarray import MCA

    Initialize standardized MCA and compute the first 2 modes of two
    ``xr.DataArray`` ``da1`` and ``da2``:

    >>> model = MCA(da1, da2, norm=True, n_modes=2)
    >>> model.solve()

    Get squared covariance fraction:

    >>> model.squared_covariance_fraction()

    Get singular vectors:

    >>> left_vectors, right_vectors = model.singular_vectors()

    Get PCs:

    >>> left_pcs, right_pcs = model.pcs()

    '''

    def __init__(
        self,
        X: DataArrayList,
        Y: DataArrayList,
        dim : Optional[Union[str, List[str]]] = 'time',
        n_modes : Optional[int] = None,
        norm : bool = False,
        weights_X : Optional[Union[str, DataArrayList]] = None,
        weights_Y : Optional[Union[str, DataArrayList]] = None
    ):
        use_coslat_X = True if weights_X == 'coslat' else False
        use_coslat_Y = True if weights_Y == 'coslat' else False

        self._tfx = _MultiDataArrayTransformer()
        self._tfy = _MultiDataArrayTransformer()

        X = self._tfx.fit_transform(X, dim=dim, coslat=use_coslat_X)
        Y = self._tfy.fit_transform(Y, dim=dim, coslat=use_coslat_Y)

        if use_coslat_X:
            weights_X = self._tfx.coslat_weights
        if use_coslat_Y:
            weights_Y = self._tfy.coslat_weights

        weights_X = self._tfx.transform_weights(weights_X)
        weights_Y = self._tfy.transform_weights(weights_Y)

        super().__init__(
            X=X,
            Y=Y,
            n_modes=n_modes,
            norm=norm,
            weights_X=weights_X,
            weights_Y=weights_Y
        )
        self._idx_mode = pd.Index(range(1, self.n_modes + 1), name='mode')

    def singular_values(self) -> DataArray:
        svalues = super().singular_values()
        svalues = xr.DataArray(
            svalues,
            dims=['mode'],
            coords=dict(mode=self._idx_mode),
            name='singular_values'
        )
        return svalues

    def explained_covariance(self) -> DataArray:
        expvar = super().explained_covariance()
        expvar = xr.DataArray(
            expvar,
            dims=['mode'],
            coords=dict(mode=self._idx_mode),
            name='explained_covariance'
        )
        return expvar

    def squared_covariance_fraction(self) -> DataArray:
        scf = super().squared_covariance_fraction()
        scf = xr.DataArray(
            scf,
            dims=['mode'],
            coords=dict(mode=self._idx_mode),
            name='squared_covariance_fraction'
        )
        return scf

    def singular_vectors(
            self, scaling : int = 0
    ) -> Tuple[DataArrayList, DataArrayList]:
        Vx, Vy = super().singular_vectors(scaling=scaling)
        Vx = self._tfx.back_transform_eofs(Vx)
        Vy = self._tfy.back_transform_eofs(Vy)
        # Rename DataArrays
        for da in Vx:
            da.name = 'left_singular_vectors'
        for da in Vy:
            da.name = 'right_singular_vectors'
        return squeeze(Vx), squeeze(Vy)

    def pcs(self, scaling : int = 0) -> Tuple[DataArray, DataArray]:
        Ux, Uy = super().pcs(scaling=scaling)
        Ux = self._tfx.back_transform_pcs(Ux)
        Uy = self._tfy.back_transform_pcs(Uy)
        # Rename DataArrays
        Ux.name = 'left_pcs'
        Uy.name = 'right_pcs'
        return Ux, Uy

    def homogeneous_patterns(self) -> Tuple[DataArrayList, DataArrayList]:
        hom_pats, pvals = super().homogeneous_patterns()
        hom_patsx = self._tfx.back_transform_eofs(hom_pats[0])
        pvalsx = self._tfx.back_transform_eofs(pvals[0])
        hom_patsy = self._tfy.back_transform_eofs(hom_pats[1])
        pvalsy = self._tfy.back_transform_eofs(pvals[1])

        # Rename DataArrays
        for da, p in zip(hom_patsx, pvalsx):
            da.name = 'left_homogeneous_patterns'
            p.name = 'left_homogeneous_patterns_p_values'
        for da, p in zip(hom_patsy, pvalsy):
            da.name = 'right_homogeneous_patterns'
            p.name = 'right_homogeneous_patterns_p_values'

        # Remove list if single DataArray
        hom_patsx = squeeze(hom_patsx)
        hom_patsy = squeeze(hom_patsy)
        pvalsx = squeeze(pvalsx)
        pvalsy = squeeze(pvalsy)
        return (hom_patsx, hom_patsy), (pvalsx, pvalsy)

    def heterogeneous_patterns(self) -> Tuple[DataArrayList, DataArrayList]:
        het_pats, pvals = super().heterogeneous_patterns()
        het_patsx = self._tfx.back_transform_eofs(het_pats[0])
        pvalsx = self._tfx.back_transform_eofs(pvals[0])
        het_patsy = self._tfy.back_transform_eofs(het_pats[1])
        pvalsy = self._tfy.back_transform_eofs(pvals[1])

        # Rename DataArrays
        for da, p in zip(het_patsx, pvalsx):
            da.name = 'left_heterogeneous_patterns'
            p.name = 'left_heterogeneous_patterns_p_values'
        for da, p in zip(het_patsy, pvalsy):
            da.name = 'right_heterogeneous_patterns'
            p.name = 'right_heterogeneous_patterns_p_values'

        # Remove list if single DataArray
        het_patsx = squeeze(het_patsx)
        het_patsy = squeeze(het_patsy)
        pvalsx = squeeze(pvalsx)
        pvalsy = squeeze(pvalsy)
        return (het_patsx, het_patsy), (pvalsx, pvalsy)

    def reconstruct_XY(
        self,
        mode : Optional[Union[int, List[int], slice]] = None
    ) -> Tuple[DataArrayList, DataArrayList]:
        Xrec, Yrec = super().reconstruct_XY(mode)
        Xrec = self._tfx.back_transform(Xrec)
        Yrec = self._tfy.back_transform(Yrec)
        return squeeze(Xrec), squeeze(Yrec)

    def project_onto_left_singular_vectors(
        self,
        X : DataArrayList = None,
        scaling : int = 0
    ) -> DataArray:
        # Transform data to 2D
        projx = _MultiDataArrayTransformer()
        X_proj = projx.fit_transform(X, dim=self._tfy.dims_samples)
        # Perform projection
        pcs_X = super().project_onto_left_singular_vectors(
            X=X_proj, scaling=scaling
        )
        # Backtransform to PC format
        pcs_X = projx.back_transform_pcs(pcs_X)
        return pcs_X

    def project_onto_right_singular_vectors(
        self,
        Y : DataArrayList = None,
        scaling : int = 0
    ) -> DataArray:
        # Transform data to 2D
        projy = _MultiDataArrayTransformer()
        Y_proj = projy.fit_transform(Y, dim=self._tfy.dims_samples)
        # Perform projection
        pcs_Y = super().project_onto_right_singular_vectors(
            Y=Y_proj, scaling=scaling
        )
        # Backtransform to PC format
        pcs_Y = projy.back_transform_pcs(pcs_Y)
        return pcs_Y



from typing import Union, Optional, Dict, Any, Tuple, List

import numpy as np
import warnings
from tqdm import trange

from ._base_eof import _BaseEOF

Array = np.ndarray
ArrayList = Union[Array, List[Array]]


class _BaseBootstrapper:
    '''Bootstrap a model to obtain significant modes and confidence intervals.

    '''

    def __init__(self, n_boot: int, alpha : float = 0.05, test_type : Optional[str] = 'two-sided'):
        '''Initialize a Bootstrapper instance.

        Parameters
        ----------
        n_boot : int
            Number of bootstraps.
        alpha : float
            Level of significance (the default is 0.05).
        test_type : ['one-sided', 'two-sided']
            Whether to perfrom one-sided or two-sided significance test.
            If two-sided, the pvalues deemed significant are pvalue < alpha / 2.
            The default is 'two-sided'.

        '''
        self._params = {
            'n_boot': n_boot,
            'alpha': alpha,
            'test_type': test_type,
        }
        if test_type == 'one-sided':
            pvalue = alpha
        elif test_type == 'two-sided':
            pvalue = alpha / 2
        else:
            raise ValueError('{:} is not a valid test_type'.format(test_type))

        n_boot_min = round(1. / pvalue)
        if n_boot_min > n_boot:
            msg = (
                'To reach a significance level of {:} '
                'use at least {:} bootstraps'
            )
            msg = msg.format(alpha, n_boot_min)
            warnings.warn(msg)
        self._params['quantiles'] = [0. + pvalue, 1. - pvalue]

    def bootstrap(self, model : _BaseEOF) -> None:
        '''Bootstrap a given model.

        Parameters
        ----------
        model : _BaseEOF
            A EOF analysis model.

        '''
        self._model = model
        n_boot = self._params['n_boot']
        n_samples = model.n_samples
        n_modes = model.n_modes
        shape_eofs = model._eofs.shape
        shape_pcs = model._pcs.shape

        self._explained_variance = np.zeros((n_boot, n_modes)) * np.nan
        self._eofs = np.zeros((n_boot,) + shape_eofs) * np.nan
        self._pcs = np.zeros((n_boot,) + shape_pcs) * np.nan

        for i in trange(n_boot, desc='Bootstrap'):
            idx_rnd = np.random.choice(n_samples, n_samples, replace=True)
            X_rnd = self._model.X[idx_rnd]
            bs_pca = _BaseEOF(X_rnd, n_modes=n_modes)
            bs_pca.solve()
            self._explained_variance[i] = bs_pca.explained_variance()
            self._eofs[i] = bs_pca.eofs()
            # Project original data onto bootstrap EOFs to obtain bootstrap PCs
            self._pcs[i] = bs_pca.project_onto_eofs(self._model.X)

        # Fix sign of individual EOFs determined by correlation coefficients
        # for a given mode with all the individual bootstrap members
        for mode in range(n_modes):
            corr_mat = np.corrcoef(self._model._eofs[:, mode], self._eofs[..., mode])
            corr_coefs = corr_mat[0][1:]
            signs = np.sign(corr_coefs)
            self._eofs[:, :, mode] = self._eofs[:, :, mode] * signs[..., None]
            self._pcs[:, :, mode] = self._pcs[:, :, mode] * signs[..., None]

        # return None
        # Extract quantiles of all quantities
        # (eigenvalues/exp. var., eigenvectors/EOFs, projections/PCs)
        self._explained_variance = np.quantile(
            self._explained_variance, self._params['quantiles'], axis=0
        )
        self._eofs = np.quantile(self._eofs, self._params['quantiles'], axis=0)
        self._pcs = np.quantile(self._pcs, self._params['quantiles'], axis=0)

        # Determine which modes are significant
        # If the lower quantile of explained variance for a given mode
        # is lower than the upper quantile of a subsquent mode, the first mode
        # is considered not significant.
        quantile_lower, quantile_upper = self._explained_variance
        is_not_sig = (quantile_lower - np.roll(quantile_upper, -1)) < 0
        self._is_significant_mode = is_not_sig.cumsum()[:-1] == 0

        # Determine which elements of the EOFs/eigenvectors are significant
        # If for a given element of the eigenvectors the quantiles cross zero,
        # the given element can not be discerned from zero ==> insignificant
        # We can easily check this by testing whether the sign of both
        # quantiles is the same (i.e. significant) or not (i.e. not significant)
        q_low, q_up = self._eofs
        self._is_significant_eof_element = np.sign(q_low) == np.sign(q_up)

        # Determine which elements of the PCs are significant
        q_low, q_up = self._pcs
        self._is_significant_pc_element = np.sign(q_low) == np.sign(q_up)

    def get_params(self) -> Dict[str, Any]:
        '''Get parameters used for bootstrapping.

        Returns
        -------
        Dict[str, Any]
            parameters used for bootstrapping.

        '''
        return self._params

    def n_significant_modes(self) -> int:
        '''Get number of significant modes.

        Returns
        -------
        int
            Number of signifcant modes
        '''
        return self._is_significant_mode.sum()

    def explained_variance(self) -> Tuple[Array, Array]:
        '''Get bootstrapped explained variance.

        Returns
        -------
        Tuple[Array, Array]
            Confidence interval of explained variance and mask
            indicating significant (True) and non-signficant (False) modes.

        '''
        return self._explained_variance, self._is_significant_mode

    def eofs(self) -> Tuple[Array, Array]:
        '''Get bootstrapped EOFs.

        Returns
        -------
        Tuple[Array, Array]
            Confidence interval of EOFs and mask
            indicating significant (True) and non-signficant (False) modes.

        '''
        return self._eofs, self._is_significant_eof_element

    def pcs(self) -> Tuple[Array, Array]:
        '''Get boostrapped PCs.

        Returns
        -------
        Tuple[Array, Array]
            Confidence interval of PCs and mask
            indicating significant (True) and non-signficant (False) elements.

        '''
        return self._pcs, self._is_significant_pc_element



from typing import Optional, Tuple, Union, List

import pandas as pd

from ..models._base_eof import _BaseEOF
from ..utils.tools import squeeze
from xeofs.pandas._transformer import _MultiDataFrameTransformer

DataFrame = pd.DataFrame
DataFrameList = Union[DataFrame, List[DataFrame]]


class EOF(_BaseEOF):
    '''EOF analysis of a single or multiple ``pd.DataFrame``.

    Parameters
    ----------
    X : pd.DataFrame
        Data to be decpomposed.
    axis : int
        Axis along which variance should be maximsed (the default is 0).
    n_modes : Optional[int]
        Number of modes to compute. Computing less modes can results in
        performance gains. If None, then the maximum number of modes is
        equivalent to ``min(n_samples, n_features)`` (the default is None).
    norm : bool
        Normalize each feature (e.g. grid cell) by its temporal standard
        deviation (the default is False).


    Examples
    --------

    Import package and create data:

    >>> import pandas as pd
    >>> from xeofs.pandas import EOF
    >>> rng = np.random.default_rng(7)
    >>> X = rng.standard_normal((4, 3))
    >>> df = pd.DataFrame(X)

    Initialize standardized EOF analysis and compute the first 2 modes:

    >>> model = EOF(df, norm=True, n_modes=2)
    >>> model.solve()

    Get explained variance:

    >>> model.explained_variance()
    ...     	explained_variance
    ...     mode
    ...     1	2.562701
    ...     2	1.167054

    Get EOFs:

    >>> model.eofs()
    ... mode	1	2
    ... 0	0.626041	-0.428431
    ... 1	0.677121	-0.115737
    ... 2	0.386755	0.896132

    Get PCs:

    >>> model.pcs()
    ... mode	1	2
    ... 0	0.495840	-0.221963
    ... 1	-2.254508	-0.470420
    ... 2	1.516900	-0.876695
    ... 3	0.241768	1.569078

    '''

    def __init__(
        self,
        X: DataFrameList,
        axis : int = 0,
        n_modes : Optional[int] = None,
        norm : bool = False,
        weights : Optional[DataFrameList] = None
    ):

        self._tf = _MultiDataFrameTransformer()
        X = self._tf.fit_transform(X, axis=axis)
        weights = self._tf.transform_weights(weights)

        super().__init__(
            X=X,
            n_modes=n_modes,
            norm=norm,
            weights=weights
        )
        self._idx_mode = pd.Index(range(1, self.n_modes + 1), name='mode')

    def singular_values(self) -> DataFrame:
        svalues = super().singular_values()
        svalues = pd.DataFrame(
            svalues,
            columns=['singular_values'],
            index=self._idx_mode
        )
        return svalues

    def explained_variance(self) -> DataFrame:
        expvar = super().explained_variance()
        expvar = pd.DataFrame(
            expvar,
            columns=['explained_variance'],
            index=self._idx_mode
        )
        return expvar

    def explained_variance_ratio(self) -> DataFrame:
        expvar = super().explained_variance_ratio()
        expvar = pd.DataFrame(
            expvar,
            columns=['explained_variance_ratio'],
            index=self._idx_mode
        )
        return expvar

    def eofs(self, scaling : int = 0) -> DataFrameList:
        eofs = super().eofs(scaling=scaling)
        eofs = self._tf.back_transform_eofs(eofs)
        return squeeze(eofs)

    def pcs(self, scaling : int = 0) -> DataFrame:
        pcs = super().pcs(scaling=scaling)
        pcs = self._tf.back_transform_pcs(pcs)
        return pcs

    def eofs_as_correlation(self) -> Tuple[DataFrameList, DataFrameList]:
        corr, pvals = super().eofs_as_correlation()
        corr = self._tf.back_transform_eofs(corr)
        pvals = self._tf.back_transform_eofs(pvals)
        return squeeze(corr), squeeze(pvals)

    def reconstruct_X(
        self,
        mode : Optional[Union[int, List[int], slice]] = None
    ) -> DataFrameList:
        Xrec = super().reconstruct_X(mode)
        Xrec = self._tf.back_transform(Xrec)
        return squeeze(Xrec)

    def project_onto_eofs(
        self,
        X : DataFrameList,
        scaling : int = 0
    ) -> DataFrame:
        '''Project new data onto the EOFs.

        Parameters
        ----------
        X : pd.DataFrame
             New data to project. Data must have same feature shape as original
             data.
        scaling : [0, 1, 2]
            Projections are scaled (i) to be orthonormal (``scaling=0``), (ii) by the
            square root of the eigenvalues (``scaling=1``) or (iii) by the
            singular values (``scaling=2``). In case no weights were applied,
            scaling by the singular values results in the projections having the
            unit of the input data (the default is 0).

        '''
        proj = _MultiDataFrameTransformer()
        X = proj.fit_transform(X, axis=self._tf.axis_samples)
        pcs = super().project_onto_eofs(X=X, scaling=scaling)
        return proj.back_transform_pcs(pcs)



from typing import Optional, Tuple, List, Union

import pandas as pd

from ..models._base_rock_pca import _BaseROCK_PCA
from ..utils.tools import squeeze
from xeofs.pandas._transformer import _MultiDataFrameTransformer

DataFrame = pd.DataFrame
DataFrameList = Union[DataFrame, List[DataFrame]]


class ROCK_PCA(_BaseROCK_PCA):
    '''ROCK-PCA of a single or multiple ``pd.DataFrame``.'''

    def __init__(
        self,
        X: DataFrame,
        n_rot : int,
        power : int,
        sigma : float,
        axis : int = 0,
        n_modes: Optional[int] = None,
        norm: bool = False,
        weights: Optional[DataFrame] = None,
    ):
        self._tf = _MultiDataFrameTransformer()
        X = self._tf.fit_transform(X)
        weights = self._tf.transform_weights(weights)

        super().__init__(
            X=X, n_rot=n_rot, power=power, sigma=sigma, n_modes=n_modes,
            norm=norm, weights=weights
        )
        self._idx_mode = pd.Index(range(1, self._params['n_modes'] + 1), name='mode')

    def explained_variance(self) -> DataFrame:
        expvar = super().explained_variance()
        return pd.DataFrame(
            expvar,
            columns=['explained_variance'],
            index=self._idx_mode
        )

    def explained_variance_ratio(self) -> DataFrame:
        expvar_ratio = super().explained_variance_ratio()
        return pd.DataFrame(
            expvar_ratio,
            columns=['explained_variance_ratio'],
            index=self._idx_mode
        )

    def eofs(self) -> DataFrameList:
        eofs = super().eofs()
        eofs = self._tf.back_transform_eofs(eofs)
        return squeeze(eofs)

    def pcs(self) -> DataFrame:
        pcs = super().pcs()
        return self._tf.back_transform_pcs(pcs)

    def eofs_amplitude(self) -> DataFrameList:
        amp = super().eofs_amplitude()
        amp = self._tf.back_transform_eofs(amp)
        return squeeze(amp)

    def pcs_amplitude(self) -> DataFrame:
        amp = super().pcs_amplitude()
        return self._tf.back_transform_pcs(amp)

    def eofs_phase(self) -> DataFrameList:
        phase = super().eofs_phase()
        phase = self._tf.back_transform_eofs(phase)
        return squeeze(phase)

    def pcs_phase(self) -> DataFrame:
        phase = super().pcs_phase()
        return self._tf.back_transform_pcs(phase)



from typing import Iterable, Optional, Union, Tuple, List

import xarray as xr

from ..models._base_eof import _BaseEOF
from ..utils.tools import squeeze
from ._transformer import _MultiDataArrayTransformer

DataArray = xr.DataArray
DataArrayList = Union[DataArray, List[DataArray]]


class EOF(_BaseEOF):
    '''EOF analysis of a single or multiple ``xr.DataArray``.

    Parameters
    ----------
    X : xr.DataArray
        Data to be decomposed.
    dim : Union[str, Iterable[str]]
        Define the dimension which should considered for maximising variance.
        For most applications in climate science, temporal variance is
        maximised (also known as S-mode EOF analysis) i.e. the time dimension
        should be chosen. If spatial variance should be maximised
        (i.e. T-mode EOF analysis), set e.g. ``dim=['lon', 'lat']``
        (the default is ``time``).
    n_modes : Union[int | None]
        Number of modes to compute. Computing less modes can results in
        performance gains. If None, then the maximum number of modes is
        equivalent to ``min(n_samples, n_features)`` (the default is None).
    norm : bool
        Normalize each feature (e.g. grid cell) by its temporal standard
        deviation (the default is False).
    weights : Union[xr.DatArray | str | None]
        Weights to be applied to data (features).


    Examples
    --------

    Import package and create data:

    >>> import xarray as xr
    >>> from xeofs.xarray import EOF
    >>> da = xr.tutorial.load_dataset('rasm')['Tair']
    >>> da = xr.tutorial.load_dataset('air_temperature')['air']
    >>> da = da.isel(lon=slice(0, 3), lat=slice(0, 2))

    Initialize standardized EOF analysis and compute the first 2 modes:

    >>> model = EOF(da, norm=True, n_modes=2)
    >>> model.solve()

    Get explained variance:

    >>> model.explained_variance()
    ... xarray.DataArray'explained_variance'mode: 2
    ...     array([5.8630486 , 0.12335653], dtype=float32)
    ... Coordinates:
    ...     mode (mode) int64 1 2
    ... Attributes:
    ...     (0)

    Get EOFs:

    >>> model.eofs()
    ... xarray.DataArray 'EOFs' lon: 3 lat: 2 mode: 2
    ... array([[[ 0.4083837 , -0.39021498],
    ...         [ 0.40758175,  0.42474997]],
    ...
    ...        [[ 0.40875185, -0.40969774],
    ...         [ 0.40863484,  0.41475943]],
    ...
    ...        [[ 0.40776297, -0.4237887 ],
    ...         [ 0.40837321,  0.38450614]]])
    ... Coordinates:
    ...     lat (lat) float32 75.0 72.5
    ...     lon (lon) float32 200.0 202.5 205.0
    ...     mode (mode) int64 1 2
    ... Attributes:
    ...     (0)

    Get PCs:

    >>> model.pcs()
    ... xarray.DataArray 'PCs' time: 2920 mode: 2
    ... array([[-3.782707  , -0.07754549],
    ...        [-3.7966802 , -0.13775176],
    ...        [-3.7969239 , -0.05770111],
    ...        ...,
    ...        [-3.2584608 ,  0.3592216 ],
    ...        [-3.031799  ,  0.2055658 ],
    ...        [-3.0840495 ,  0.25031802]], dtype=float32)
    ... Coordinates:
    ...     time (time) datetime64[ns] 2013-01-01 ... 2014-12-31T18:00:00
    ...     mode (mode) int64 1 2
    ... Attributes:
    ...     (0)

    '''

    def __init__(
        self,
        X: DataArrayList,
        dim: Union[str, Iterable[str]] = 'time',
        n_modes : Optional[int] = None,
        norm : bool = False,
        weights : Optional[Union[xr.DataArray, str]] = None
    ):
        use_coslat = True if weights == 'coslat' else False

        self._tf = _MultiDataArrayTransformer()
        X = self._tf.fit_transform(X, dim=dim, coslat=use_coslat)

        if use_coslat:
            weights = self._tf.coslat_weights
        weights = self._tf.transform_weights(weights)

        super().__init__(
            X=X,
            n_modes=n_modes,
            norm=norm,
            weights=weights
        )
        self._idx_mode = xr.IndexVariable('mode', range(1, self.n_modes + 1))
        self._dim = dim

    def singular_values(self) -> DataArray:
        svalues = super().singular_values()
        return xr.DataArray(
            svalues,
            dims=['mode'],
            coords={'mode' : self._idx_mode},
            name='singular_values'
        )

    def explained_variance(self) -> DataArray:
        expvar = super().explained_variance()
        return xr.DataArray(
            expvar,
            dims=['mode'],
            coords={'mode' : self._idx_mode},
            name='explained_variance'
        )

    def explained_variance_ratio(self) -> DataArray:
        expvar = super().explained_variance_ratio()
        return xr.DataArray(
            expvar,
            dims=['mode'],
            coords={'mode' : self._idx_mode},
            name='explained_variance_ratio'
        )

    def eofs(self, scaling : int = 0) -> DataArrayList:
        eofs = super().eofs(scaling=scaling)
        eofs = self._tf.back_transform_eofs(eofs)
        return squeeze(eofs)

    def pcs(self, scaling : int = 0) -> xr.DataArray:
        pcs = super().pcs(scaling=scaling)
        pcs = self._tf.back_transform_pcs(pcs)
        return pcs

    def eofs_as_correlation(self) -> Tuple[DataArrayList, DataArrayList]:
        corr, pvals = super().eofs_as_correlation()
        corr = self._tf.back_transform_eofs(corr)
        pvals = self._tf.back_transform_eofs(pvals)
        for c, p in zip(corr, pvals):
            c.name = 'correlation_coeffient'
            p.name = 'p_value'
        return squeeze(corr), squeeze(pvals)

    def reconstruct_X(
        self,
        mode : Optional[Union[int, List[int], slice]] = None
    ) -> DataArrayList:
        Xrec = super().reconstruct_X(mode=mode)
        Xrec = self._tf.back_transform(Xrec)
        return squeeze(Xrec)

    def project_onto_eofs(
        self,
        X : DataArrayList,
        scaling : int = 0
    ) -> DataArray:
        '''Project new data onto the EOFs.

        Parameters
        ----------
        X : xr.DataArray
             New data to project. Data must have same feature shape as original
             data.
        scaling : [0, 1, 2]
            Projections are scaled (i) to be orthonormal (``scaling=0``), (ii) by the
            square root of the eigenvalues (``scaling=1``) or (iii) by the
            singular values (``scaling=2``). In case no weights were applied,
            scaling by the singular values results in the projections having the
            unit of the input data (the default is 0).

        '''
        proj = _MultiDataArrayTransformer()
        X = proj.fit_transform(X, dim=self._tf.dims_samples)
        pcs = super().project_onto_eofs(X=X, scaling=scaling)
        return proj.back_transform_pcs(pcs)



from typing import Union, Iterable, List
import numpy as np
import xarray as xr

from ..models._transformer import _ArrayTransformer, _MultiArrayTransformer


class _DataArrayTransformer(_ArrayTransformer):
    '''`_ArrayTransformer` wrapper for `xarray.DataArray`.

    '''

    def __init__(self):
        super().__init__()

    def fit(
        self,
        X : xr.DataArray,
        dim : Union[str, Iterable[str]] = 'time',
        coslat : bool = False
    ):
        if not isinstance(X, xr.DataArray):
            raise ValueError('This interface is for `xarray.DataArray` only')
        # Ensure that dims are tuple
        if isinstance(dim, str):
            dim = [dim]
        dim = tuple(dim)
        # Get axis position for each dim
        if np.isin(dim, X.dims).all():
            axis = [np.argmax(np.isin(X.dims, d)) for d in dim]
        else:
            err_msg = 'One or more of {:} is not a valid dimension.'
            err_msg = err_msg.format(dim)
            raise ValueError(err_msg)

        self.dims = X.dims
        self.dims_samples = dim
        self.dims_features = tuple(d for d in self.dims if d not in dim)

        self.coords = X.coords
        self.coords_samples = {
            k: c for k, c in self.coords.items() if k in self.dims_samples
        }
        self.coords_features = {
            k: c for k, c in self.coords.items() if k in self.dims_features
        }
        if coslat:
            self.coslat_weights = self._get_coslat_weights(X)
        else:
            self.coslat_weights = None

        super().fit(X=X.values, axis=axis)
        return self

    def _get_coslat_weights(self, X : xr.DataArray):
        # Find dimension name of latitude
        possible_lat_names = [
            'latitude', 'Latitude', 'lat', 'Lat', 'LATITUDE', 'LAT'
        ]
        idx_lat_dim = np.isin(X.dims, possible_lat_names)
        try:
            lat_dim = np.array(X.dims)[idx_lat_dim][0]
        except IndexError:
            err_msg = (
                'Latitude dimension cannot be found. Please make sure '
                'latitude dimensions is called like one of {:}'
            )
            err_msg = err_msg.format(possible_lat_names)
            raise ValueError(err_msg)
        # Check if latitude is a MultiIndex => not allowed
        if X.coords[lat_dim].dtype not in [np.float_, np.float64, np.float32, np.int_]:
            err_msg = 'MultiIndex as latitude dimensions is not allowed.'
            raise ValueError(err_msg)
        # Compute coslat weights
        weights = np.cos(np.deg2rad(X.coords[lat_dim]))
        weights = np.sqrt(weights.where(weights > 0, 0))
        # Broadcast latitude weights on other feature dimensions
        sample_dims = self.dims_samples
        feature_grid = X.isel({k: 0 for k in sample_dims})
        feature_grid = feature_grid.drop_vars(sample_dims)
        return weights.broadcast_like(feature_grid)

    def transform(self, X : xr.DataArray):
        dims_expected = self.dims
        try:
            dims_received = X.dims
        except AttributeError:
            err_msg = 'X must be of type {:}.'.format(repr(xr.DataArray))
            raise TypeError(err_msg)
        if dims_expected == dims_received:
            return super().transform(X.values)
        else:
            err_msg = 'Expected dimensions {:}, but got {:} instead.'
            err_msg = err_msg.format(dims_expected, dims_received)
            raise ValueError(err_msg)

    def fit_transform(
        self,
        X : xr.DataArray,
        dim : Union[str, Iterable[str]] = 'time',
        coslat : bool = False
    ):
        return self.fit(X=X, dim=dim, coslat=coslat).transform(X=X)

    def transform_weights(self, weights : xr.DataArray):
        dims_expected = self.dims_features
        try:
            dims_received = weights.dims
        except AttributeError:
            return super().transform_weights(weights)
        if dims_expected == dims_received:
            return super().transform_weights(weights.values)
        else:
            err_msg = 'Expected dimensions {:}, but got {:} instead.'
            err_msg = err_msg.format(dims_expected, dims_received)
            raise ValueError(err_msg)

    def back_transform(self, X : np.ndarray):
        da = super().back_transform(X)
        return xr.DataArray(
            da, dims=self.dims, coords=self.coords, name='reconstructed_data'
        )

    def back_transform_eofs(self, X : np.ndarray):
        da = super().back_transform_eofs(X)
        dims = self.dims_features + ('mode',)
        coords = self.coords_features
        coords['mode'] = range(1, X.shape[1] + 1)
        return xr.DataArray(da, dims=dims, coords=coords, name='EOFs')

    def back_transform_pcs(self, X : np.ndarray):
        da = super().back_transform_pcs(X)
        dims = self.dims_samples + ('mode',)
        coords = self.coords_samples
        coords['mode'] = range(1, X.shape[1] + 1)
        return xr.DataArray(da, dims=dims, coords=coords, name='PCs')


class _MultiDataArrayTransformer(_MultiArrayTransformer):
    'Transform multiple N-D ``xr.DataArray`` to a single 2D ``np.ndarry``.'
    def __init__(self):
        super().__init__()

    def fit(
        self,
        X : Union[xr.DataArray, List[xr.DataArray]],
        dim : Union[str, Iterable[str]] = 'time',
        coslat : bool = False
    ):
        X = self._convert2list(X)
        self.tfs = [_DataArrayTransformer().fit(x, dim=dim, coslat=coslat) for x in X]

        if len(set([tf.n_valid_samples for tf in self.tfs])) > 1:
            err_msg = 'All individual arrays must have same number of samples.'
            raise ValueError(err_msg)

        self.idx_array_sep = np.cumsum([tf.n_valid_features for tf in self.tfs])
        self.dims_samples = self.tfs[0].dims_samples
        self.coslat_weights = [tf.coslat_weights for tf in self.tfs]
        return self

    def transform(self, X : Union[xr.DataArray, List[xr.DataArray]]) -> np.ndarray:
        return super().transform(X=X)

    def transform_weights(self, weights : Union[xr.DataArray, List[xr.DataArray]]) -> np.ndarray:
        return super().transform_weights(weights=weights)

    def fit_transform(
        self, X : Union[xr.DataArray, List[xr.DataArray]],
        dim : Union[str, Iterable[str]] = 'time',
        coslat : bool = False
    ) -> np.ndarray:
        return self.fit(X=X, dim=dim, coslat=coslat).transform(X)

    def back_transform(self, X : np.ndarray) -> List[xr.DataArray]:
        return super().back_transform(X=X)

    def back_transform_eofs(self, X : np.ndarray) -> List[xr.DataArray]:
        return super().back_transform_eofs(X=X)

    def back_transform_pcs(self, X : np.ndarray) -> List[xr.DataArray]:
        return super().back_transform_pcs(X=X)



from typing import Union, Iterable, List
import numpy as np
import pandas as pd

from ..models._transformer import _ArrayTransformer, _MultiArrayTransformer


class _DataFrameTransformer(_ArrayTransformer):
    '''`_ArrayTransformer` wrapper for `pandas.DataFrame`.

    '''

    def __init__(self):
        super().__init__()

    def fit(self, X : pd.DataFrame, axis : Union[int, Iterable[int]] = 0):
        if not isinstance(X, pd.DataFrame):
            raise ValueError('This interface is for `pandas.DataFrame` only')
        if isinstance(axis, list):
            axis = axis[0]
        # Set sample and feature index
        if axis == 0:
            self.index_samples = X.index
            self.index_features = X.columns
        elif axis == 1:
            self.index_samples = X.columns
            self.index_features = X.index
        else:
            raise ValueError('axis must be either 0 or 1')
        # Fit the data
        try:
            super().fit(X=X.values, axis=axis)
        except AttributeError:
            err_msg = 'weights must be of type {:}.'.format(repr(pd.DataFrame))
            raise TypeError(err_msg)
        return self

    def transform(self, X : pd.DataFrame) -> np.ndarray:
        try:
            return super().transform(X.values)
        except AttributeError:
            err_msg = 'weights must be of type {:}.'.format(repr(pd.DataFrame))
            raise TypeError(err_msg)

    def fit_transform(self, X : pd.DataFrame, axis : int = 0) -> np.ndarray:
        return self.fit(X=X, axis=axis).transform(X)

    def transform_weights(self, weights : pd.DataFrame) -> np.ndarray:
        try:
            return super().transform_weights(weights.values)
        except AttributeError:
            return super().transform_weights(weights)

    def back_transform(self, X : np.ndarray) -> pd.DataFrame:
        df = super().back_transform(X)
        return pd.DataFrame(
            df,
            index=self.index_samples,
            columns=self.index_features
        )

    def back_transform_eofs(self, X : np.ndarray) -> pd.DataFrame:
        eofs = super().back_transform_eofs(X)
        return pd.DataFrame(
            eofs,
            index=self.index_features,
            columns=range(1, eofs.shape[-1] + 1)
        )

    def back_transform_pcs(self, X : np.ndarray) -> pd.DataFrame:
        pcs = super().back_transform_pcs(X)
        return pd.DataFrame(
            pcs,
            index=self.index_samples,
            columns=range(1, pcs.shape[-1] + 1)
        )


class _MultiDataFrameTransformer(_MultiArrayTransformer):
    'Transform multiple 2D ``pd.DataFrame`` to a single 2D ``np.ndarry``.'
    def __init__(self):
        super().__init__()

    def fit(self, X : Union[pd.DataFrame, List[pd.DataFrame]], axis : Union[int, Iterable[int]] = 0):
        X = self._convert2list(X)
        self.tfs = [_DataFrameTransformer().fit(x, axis=axis) for x in X]

        if len(set([tf.n_valid_samples for tf in self.tfs])) > 1:
            err_msg = 'All individual arrays must have same number of samples.'
            raise ValueError(err_msg)

        self.idx_array_sep = np.cumsum([tf.n_valid_features for tf in self.tfs])
        self.axis_samples = self.tfs[0].axis_samples
        return self

    def transform(self, X : Union[pd.DataFrame, List[pd.DataFrame]]) -> np.ndarray:
        return super().transform(X=X)

    def transform_weights(self, weights : Union[pd.DataFrame, List[pd.DataFrame]]) -> np.ndarray:
        return super().transform_weights(weights=weights)

    def fit_transform(
        self, X : Union[pd.DataFrame, List[pd.DataFrame]],
        axis : Union[int, Iterable[int]] = 0
    ) -> np.ndarray:
        return self.fit(X=X, axis=axis).transform(X)

    def back_transform(self, X : np.ndarray) -> pd.DataFrame:
        return super().back_transform(X=X)

    def back_transform_eofs(self, X : np.ndarray) -> pd.DataFrame:
        return super().back_transform_eofs(X=X)

    def back_transform_pcs(self, X : np.ndarray) -> pd.DataFrame:
        return super().back_transform_pcs(X=X)



import numpy as np
from typing import Optional, Union, List, Tuple

from ._base_mca_rotator import _BaseMCARotator
from ._transformer import _MultiArrayTransformer
from ..utils.tools import squeeze

Array = np.ndarray
ArrayList = Union[Array, List[Array]]


class MCA_Rotator(_BaseMCARotator):
    '''Rotates a solution obtained from ``xe.models.MCA``.'''

    def __init__(
        self,
        n_rot : int,
        loadings : str = 'standard',
        power : int = 1,
        max_iter : int = 1000,
        rtol : float = 1e-8
    ):

        super().__init__(
            n_rot=n_rot, loadings=loadings, power=power,
            max_iter=max_iter, rtol=rtol
        )

    def singular_values(self) -> Array:
        return super().singular_values()

    def explained_covariance(self) -> Array:
        return super().explained_covariance()

    def squared_covariance_fraction(self) -> Array:
        return super().squared_covariance_fraction()

    def singular_vectors(self, scaling : int = 0) -> ArrayList:
        Vx, Vy = super().singular_vectors(scaling=scaling)
        Vx = self._model._tfx.back_transform_eofs(Vx)
        Vy = self._model._tfy.back_transform_eofs(Vy)
        return squeeze(Vx), squeeze(Vy)

    def pcs(self, scaling : int = 0) -> Array:
        Ux, Uy = super().pcs(scaling=scaling)
        Ux = self._model._tfx.back_transform_pcs(Ux)
        Uy = self._model._tfy.back_transform_pcs(Uy)
        return Ux, Uy

    def homogeneous_patterns(self) -> Tuple[ArrayList, ArrayList]:
        hom_pats, pvals = super().homogeneous_patterns()
        hom_pats_x = self._model._tfx.back_transform_eofs(hom_pats[0])
        hom_pats_y = self._model._tfy.back_transform_eofs(hom_pats[1])
        pvals_x = self._model._tfx.back_transform_eofs(pvals[0])
        pvals_y = self._model._tfy.back_transform_eofs(pvals[1])
        hom_pats_x = squeeze(hom_pats_x)
        hom_pats_y = squeeze(hom_pats_y)
        pvals_x = squeeze(pvals_x)
        pvals_y = squeeze(pvals_y)
        return (hom_pats_x, hom_pats_y), (pvals_x, pvals_y)

    def heterogeneous_patterns(self) -> Tuple[ArrayList, ArrayList]:
        het_pats, pvals = super().heterogeneous_patterns()
        het_pats_x = self._model._tfx.back_transform_eofs(het_pats[0])
        het_pats_y = self._model._tfy.back_transform_eofs(het_pats[1])
        pvals_x = self._model._tfx.back_transform_eofs(pvals[0])
        pvals_y = self._model._tfy.back_transform_eofs(pvals[1])
        het_pats_x = squeeze(het_pats_x)
        het_pats_y = squeeze(het_pats_y)
        pvals_x = squeeze(pvals_x)
        pvals_y = squeeze(pvals_y)
        return (het_pats_x, het_pats_y), (pvals_x, pvals_y)

    def reconstruct_XY(
        self,
        mode : Optional[Union[int, List[int], slice]] = None
    ) -> ArrayList:
        Xrec, Yrec = super().reconstruct_XY(mode=mode)
        Xrec = self._model._tfx.back_transform(Xrec)
        Yrec = self._model._tfy.back_transform(Yrec)
        return squeeze(Xrec), squeeze(Yrec)

    def project_onto_left_singular_vectors(
        self,
        X : ArrayList,
        scaling : int = 0
    ) -> Array:
        '''Project new data onto the rotated EOFs.

        Parameters
        ----------
        X : np.ndarray
             New data to project onto left singular vector. Data must have
             same feature shape as original data.
        scaling : [0, 1, 2]
            Projections are scaled (i) to be orthonormal (``scaling=0``), (ii) by the
            square root of the eigenvalues (``scaling=1``) or (iii) by the
            singular values (``scaling=2``) (the default is 0).

        '''
        proj = _MultiArrayTransformer()
        X = proj.fit_transform(X, axis=self._model._tfx.axis_samples)
        pcs = super().project_onto_left_singular_vectors(X=X, scaling=scaling)
        return proj.back_transform_pcs(pcs)

    def project_onto_right_singular_vectors(
        self,
        Y : ArrayList,
        scaling : int = 0
    ) -> Array:
        '''Project new data onto the rotated EOFs.

        Parameters
        ----------
        Y : np.ndarray
             New data to project onto right singular vector. Data must have
             same feature shape as original data.
        scaling : [0, 1, 2]
            Projections are scaled (i) to be orthonormal (``scaling=0``), (ii) by the
            square root of the eigenvalues (``scaling=1``) or (iii) by the
            singular values (``scaling=2``) (the default is 0).

        '''
        proj = _MultiArrayTransformer()
        Y = proj.fit_transform(Y, axis=self._model._tfy.axis_samples)
        pcs = super().project_onto_right_singular_vectors(Y=Y, scaling=scaling)
        return proj.back_transform_pcs(pcs)



import pandas as pd
from typing import Optional, Union, List, Tuple

from .eof import EOF
from ._transformer import _MultiDataFrameTransformer
from ..utils.tools import squeeze
from ..models._base_rotator import _BaseRotator

DataFrame = pd.DataFrame
DataFrameList = Union[DataFrame, List[DataFrame]]


class Rotator(_BaseRotator):
    '''Rotates a solution obtained from ``xe.pandas.EOF``.

    Parameters
    ----------
    model : xe.pandas.EOF
        A EOF model solution.
    n_rot : int
        Number of modes to be rotated.
    power : int
        Defines the power of Promax rotation. Choosing ``power=1`` equals
        a Varimax solution (the default is 1).
    max_iter : int
        Number of maximal iterations for obtaining the rotation matrix
        (the default is 1000).
    rtol : float
        Relative tolerance to be achieved for early stopping the iteration
        process (the default is 1e-8).


    '''

    def __init__(
        self,
        model : EOF,
        n_rot : int,
        power : int = 1,
        max_iter : int = 1000,
        rtol : float = 1e-8
    ):
        super().__init__(
            model=model, n_rot=n_rot, power=power, max_iter=max_iter, rtol=rtol
        )

    def explained_variance(self) -> DataFrame:
        expvar = super().explained_variance()
        return DataFrame(
            expvar,
            index=self._model._idx_mode[:self._n_rot]
        )

    def explained_variance_ratio(self) -> DataFrame:
        expvar_ratio = super().explained_variance_ratio()
        return DataFrame(
            expvar_ratio,
            index=self._model._idx_mode[:self._n_rot]
        )

    def eofs(self, scaling : int = 0) -> DataFrameList:
        eofs = super().eofs(scaling=scaling)
        eofs = self._model._tf.back_transform_eofs(eofs)
        return squeeze(eofs)

    def pcs(self, scaling : int = 0) -> DataFrame:
        pcs = super().pcs(scaling=scaling)
        return self._model._tf.back_transform_pcs(pcs)

    def eofs_as_correlation(self) -> Tuple[DataFrameList, DataFrameList]:
        corr, pvals = super().eofs_as_correlation()
        corr = self._model._tf.back_transform_eofs(corr)
        pvals = self._model._tf.back_transform_eofs(pvals)
        return squeeze(corr), squeeze(pvals)

    def reconstruct_X(
        self,
        mode : Optional[Union[int, List[int], slice]] = None
    ) -> DataFrameList:
        Xrec = super().reconstruct_X(mode=mode)
        Xrec = self._model._tf.back_transform(Xrec)
        return squeeze(Xrec)

    def project_onto_eofs(
        self,
        X : DataFrameList,
        scaling : int = 0
    ) -> DataFrame:
        '''Project new data onto the rotated EOFs.

        Parameters
        ----------
        X : pd.DataFrame
             New data to project. Data must have same feature shape as original
             data.
        scaling : [0, 1, 2]
            Projections are scaled (i) to be orthonormal (``scaling=0``), (ii) by the
            square root of the eigenvalues (``scaling=1``) or (iii) by the
            singular values (``scaling=2``). In case no weights were applied,
            scaling by the singular values results in the projections having the
            unit of the input data (the default is 0).

        '''
        proj = _MultiDataFrameTransformer()
        X = proj.fit_transform(X, axis=self._model._tf.axis_samples)
        pcs = super().project_onto_eofs(X=X, scaling=scaling)
        return proj.back_transform_pcs(pcs)



import numpy as np
import scipy as sc
from typing import Optional, Union, List, Tuple

from .eof import EOF
from ..utils.rotation import promax
from ..utils.tools import get_mode_selector


class _BaseRotator:
    '''Rotates a solution obtained from ``xe.models.EOF``.

    Parameters
    ----------
    model : xe.models.EOF
        A EOF model solution.
    n_rot : int
        Number of modes to be rotated.
    power : int
        Defines the power of Promax rotation. Choosing ``power=1`` equals
        a Varimax solution (the default is 1).
    max_iter : int
        Number of maximal iterations for obtaining the rotation matrix
        (the default is 1000).
    rtol : float
        Relative tolerance to be achieved for early stopping the iteration
        process (the default is 1e-8).


    '''

    def __init__(
        self,
        model : EOF,
        n_rot : int,
        power : int = 1,
        max_iter : int = 1000,
        rtol : float = 1e-8
    ):
        self._power = power
        self._n_rot = n_rot
        self._model = model
        eofs = model._eofs[:, :n_rot]
        eigenvalues = model._explained_variance[:n_rot]
        L = eofs * np.sqrt(eigenvalues)
        eofs_rot, rot_mat, corr_mat = promax(
            L, power=power, max_iter=max_iter, rtol=rtol
        )
        self._rot_mat = rot_mat
        self._explained_variance = (eofs_rot ** 2).sum(axis=0)
        self._idx_var = np.argsort(self._explained_variance)[::-1]
        # Reorder rotated EOFs according to their variance
        self._eofs = eofs_rot[:, self._idx_var]
        self._explained_variance = self._explained_variance[self._idx_var]
        # Calculate explained variance fraction by using total variance
        expvar = self._model._explained_variance[0]
        frac_expvar = self._model._explained_variance_ratio[0]
        total_variance = expvar / frac_expvar
        # Calculate variance ratio
        self._explained_variance_ratio = self._explained_variance / total_variance

        # Normalize EOFs
        self._eofs = self._eofs / np.sqrt(self._explained_variance)

        # Rotate PCs using rotatation matrix
        self._pcs = self._model._pcs[:, :self._n_rot]
        R = self._rotation_matrix(inverse_transpose=True)
        self._pcs = self._pcs @ R
        # Reorder according to variance
        self._pcs = self._pcs[:, self._idx_var]

    def _rotation_matrix(self, inverse_transpose : bool = False) -> np.ndarray:
        '''Return the rotation matrix.

        Parameters
        ----------
        inverse_transpose : boolean
            If True, return the inverse transposed of the rotation matrix.
            For orthogonal rotations (Varimax) the inverse transpose equals
            the rotation matrix itself. For oblique rotations (Promax), it
            will be different in general (the default is False).

        Returns
        -------
        rotation_matrix : np.ndarray

        '''
        R = self._rot_mat

        # only for oblique rotations
        # If rotation is orthogonal: R == R^(-1).T
        # If rotation is oblique (i.e. power>1): R != R^(-1).T
        if inverse_transpose and self._power > 1:
            R = np.linalg.pinv(R).conjugate().T

        return R

    def explained_variance(self) -> np.ndarray:
        '''Explained variance after rotation.'''

        return self._explained_variance

    def explained_variance_ratio(self) -> np.ndarray:
        '''Explained variance ratio after rotation.'''

        return self._explained_variance_ratio

    def eofs(self, scaling : int = 0) -> np.ndarray:
        '''EOFs after rotation.

        Parameters
        ----------
        scaling : [0, 1, 2]
            EOFs are scaled (i) to have unit length (``scaling=0``), (ii) by the
            square root of the eigenvalues (``scaling=1``) or (iii) by the
            singular values (``scaling=2``). In case no weights were applied,
            scaling by the singular values results in the EOFs having the
            unit of the input data (the default is 0).

        '''
        dof = self._model.n_samples - 1
        if scaling == 0:
            eofs = self._eofs
        elif scaling == 1:
            eofs = self._eofs * np.sqrt(self._explained_variance)
        elif scaling == 2:
            eofs = self._eofs * np.sqrt(self._explained_variance * dof)
        return eofs

    def pcs(self, scaling : int = 0) -> np.ndarray:
        '''PCs after rotation.

        Parameters
        ----------
        scaling : [0, 1, 2]
            PCs are scaled (i) to have unit length (orthonormal for Varimax
            rotation) (``scaling=0``), (ii) by the square root of the
            eigenvalues (``scaling=1``) or (iii) by the
            singular values (``scaling=2``). In case no weights were applied,
            scaling by the singular values results in the PCs having the
            unit of the input data (the default is 0).

        '''
        dof = self._model.n_samples - 1
        if scaling == 0:
            pcs = self._pcs
        elif scaling == 1:
            pcs = self._pcs * np.sqrt(self._explained_variance)
        elif scaling == 2:
            pcs = self._pcs * np.sqrt(self._explained_variance * dof)
        return pcs

    def eofs_as_correlation(self) -> Tuple[np.ndarray, np.ndarray]:
        '''Correlation coefficients between rotated PCs and data matrix.

        Returns
        -------
        Tuple[np.ndarray, np.ndarry]
            Matrices of correlation coefficients and associated
            two-sided p-values with features as rows and modes as columns.

        '''

        # Compute correlation matrix
        corr = np.corrcoef(self._model.X, self._pcs, rowvar=False)
        corr = corr[:self._model.n_features, self._model.n_features:]
        # Compute two-sided p-values
        # Reference: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.pearsonr.html#r8c6348c62346-1
        a = self._model.n_samples / 2 - 1
        dist = sc.stats.beta(a, a, loc=-1, scale=2)
        pvals = 2 * dist.cdf(-abs(corr))
        return corr, pvals

    def reconstruct_X(
        self,
        mode : Optional[Union[int, List[int], slice]] = None
    ) -> np.ndarray:
        '''Reconstruct original data field ``X`` using the rotated PCs and EOFs.

        If weights were applied, ``X`` will be automatically rescaled.

        Parameters
        ----------
        mode : Optional[Union[int, List[int], slice]]
            Mode(s) based on which ``X`` will be reconstructed. If ``mode`` is
            an int, a single mode is used. If a list of integers is provided,
            use all specified modes for reconstruction. Alternatively, you may
            want to select a slice to reconstruct. The first mode is denoted
            by 1 (and not by 0). If None then ``X`` is recontructed using all
            available modes (the default is None).

        Examples
        --------

        Perform an analysis using some data ``X``:

        >>> model = EOF(X, norm=True)
        >>> model.solve()

        Reconstruct ``X`` using all modes:

        >>> model.reconstruct_X()

        Reconstruct ``X`` using the first mode only:

        >>> model.reconstruct_X(1)

        Reconstruct ``X`` using mode 1, 3 and 4:

        >>> model.reconstruct_X([1, 3, 4])

        Reconstruct ``X`` using all modes up to mode 10 (including):

        >>> model.reconstruct_X(slice(10))

        Reconstruct ``X`` using every second mode between 4 and 8 (both
        including):

        >>> model.reconstruct_X(slice(4, 8, 2))


        '''
        dof = self._model.n_samples - 1
        eofs = self._eofs
        pcs = self._pcs * np.sqrt(self._explained_variance * dof)
        # Select modes to reconstruct X
        mode = get_mode_selector(mode)
        eofs = eofs[:, mode]
        pcs = pcs[:, mode]
        Xrec = pcs @ eofs.T
        # Unweight and add mean
        return (Xrec / self._model._weights) + self._model._X_mean

    def project_onto_eofs(
        self,
        X : np.ndarray, scaling : int = 0
    ) -> np.ndarray:
        '''Project new data onto the rotated EOFs.

        Parameters
        ----------
        X : np.ndarray
             New data to project. Data must be a 2D matrix.
        scaling : [0, 1, 2]
            Projections are scaled (i) to be orthonormal (``scaling=0``), (ii)
            by the square root of the eigenvalues (``scaling=1``) or (iii) by
            the singular values (``scaling=2``). In the latter case, and when
            no weights were applied, scaling by the singular values results in
            the projections having the unit of the input data
            (the default is 0).

        '''
        dof = self._model.n_samples - 1
        svals = np.sqrt(self._model._explained_variance * dof)

        # Preprocess new data
        try:
            X -= self._model._X_mean
        except ValueError:
            err_msg = (
                'New data has invalid feature dimensions and cannot be '
                'projected onto EOFs.'
            )
            raise ValueError(err_msg)
        X *= self._model._weights

        # Compute non-rotated PCs
        pcs = X @ self._model._eofs[:, :self._n_rot] / svals[:self._n_rot]

        # Rotate and reorder PCs
        R = self._rotation_matrix(inverse_transpose=True)
        pcs = pcs @ R
        pcs = pcs[:, self._idx_var]

        # Apply scaling
        if scaling == 0:
            return pcs
        elif scaling == 1:
            return pcs * np.sqrt(self._explained_variance)
        elif scaling == 2:
            return pcs * np.sqrt(self._explained_variance * dof)
        else:
            err_msg = (
                'Scaling option {:} is not valid but must be one '
                'of [0, 1, 2]'
            )
            err_msg = err_msg.foramt(scaling)
            raise ValueError(err_msg)



from typing import Optional, Tuple, List, Union

import numpy as np
from scipy.signal import hilbert

from ..utils.rotation import promax
from ..utils.distance_matrix import distance_matrix


class _BaseROCK_PCA():
    '''Base class for ROCK-PCA.'''

    def __init__(
        self,
        X: np.ndarray,
        n_rot : int,
        power : int,
        sigma : float,
        n_modes: Optional[int] = None,
        norm: bool = False,
        weights: Optional[np.ndarray] = None,
    ):
        '''
        Perform rotated complex kernel PCA (ROCK-PCA).

        ROCK-PCA is a non-linear complex decomposition with an additional
        rotation in the Fourier space.

        Original reference:
        - Bueso, D., Piles, M. & Camps-Valls, G. Nonlinear PCA for Spatio-Temporal
        Analysis of Earth Observation Data. IEEE Trans. Geosci. Remote Sensing 1–12
        (2020) doi:10.1109/TGRS.2020.2969813.

        Parameters
        ----------
        X : np.ndarray
            2D data matrix ``X`` with samples as rows and features as columns.
            NaN entries will raise an error.
        n_rot : int
            Number of PCs to be rotated.
        power : int
            Power parameter for Promax rotation. If ``power=0``, no rotation is performed.
            For ``power=1``, this is equivalent to Varimax rotation.
        sigma : float,
            Hyperparamter defining the "width" Gaussian kernel
        n_modes : Optional[int]
            Number of modes to be retrieved. If None, then all possible modes will
            be computed. Reducing ``n_modes`` can greatly speed up computational
            (the default is None).
        norm : bool
            Normalize each feature by its standard deviation (the default is False).
        weights : Optional[np.ndarray]
            Apply `weights` to features. Weights must be a 1D array with the same
            length as number of features. No NaN entries are allowed
            (the default is None).

        References
        ----------
        .. [1] https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=8989964&casa_token=3zKG0dtp-ewAAAAA:FM1CrVISSSqhWEAwPGpQqCgDYccfLG4N-67xNNDzUBQmMvtIOHuC7T6X-TVQgbDg3aDOpKBksg&tag=1
        .. [2] https://github.com/DiegoBueso/ROCK-PCA

        '''
        self._params = {
            'n_rot': n_rot,
            'power': power,
            'sigma': sigma,
            'norm': norm
        }
        # Remove mean for each feature
        self._X_mean = X.mean(axis=0)
        X -= self._X_mean

        # Weights are applied to features, not samples.
        self._weights = weights
        if self._weights is None:
            # Use int type to ensure that there won't be rounding errors
            # when applying trivial weighting (= all weights equal 1)
            self._weights = np.ones(X.shape[1], dtype=int)

        # Standardization is included as weights
        if norm:
            stdev = X.std(axis=0)
            if (stdev == 0).any():
                err_msg = (
                    'Standard deviation of one ore more features is zero, '
                    'normalization not possible.'
                )
                raise ValueError(err_msg)
            self._weights = self._weights / stdev
        X = X * self._weights

        self.n_samples = X.shape[0]
        self.n_features = X.shape[1]

        if n_modes is None:
            n_modes = min(self.n_samples, self.n_features)
        self._params['n_modes'] = min(n_modes, n_rot)

        # Complexify data using Hilbert transform
        self.X = hilbert(X, axis=0)
        self.X -= self.X.mean(axis=0)

    def solve(self, algorithm='original') -> None:
        '''
        Perform complex non-linear decomposition.

        '''
        # Sigma estimation from median distance
        print('Build distance matrix ... ', flush=True)
        sigma = self._params['sigma']
        n_rot = self._params['n_rot']
        power = self._params['power']
        n_modes = self._params['n_modes']

        distance = distance_matrix(self.X.T, self.X.T, algorithm)
        print('Build kernel matrix ... ', flush=True)
        kernel = np.exp(-distance / (2 * (sigma**2)))
        center_matrix = np.identity(self.n_samples) - (1. / self.n_samples) * np.ones(self.n_samples)
        kernel = center_matrix @ kernel @ center_matrix
        print('SVD ... ', flush=True)
        U, lambdas, _ = np.linalg.svd(kernel, full_matrices=False)
        self.lambdas = lambdas

        # Rotation in frequency space
        U = U[:, :n_rot]
        if power > 0:
            print('Rotation ... ', flush=True)
            Ufft = np.fft.fft(U, axis=0)
            # Varimax
            U, _, _    = promax(Ufft, power=1)
            # Promax according to Bueso et al. implementation:
            # https://github.com/DiegoBueso/ROCK-PCA/blob/master/rock-code/Promax.m
            U = np.sign(U) * abs(U)**power
            # Renormalization
            U = abs(U) / abs(Ufft) * Ufft
            U = np.real(np.fft.ifft(U, axis=0))
            U = hilbert(U, axis=0)
            U = center_matrix @ U

        # Estimate variance explained
        self._explained_variance = np.linalg.norm(U.conj().T @ self.X, axis=1) / np.linalg.norm(U, axis=0)
        self._total_variance = self._explained_variance.sum()
        self._explained_variance_ratio = self._explained_variance / self._total_variance

        # Reorder such that first mode explained most variance
        idx_sort = np.argsort(self._explained_variance)[::-1]
        self._explained_variance = self._explained_variance[idx_sort]
        self._explained_variance_ratio = self._explained_variance_ratio[idx_sort]
        self._pcs = U[:, idx_sort]

        self._explained_variance = self._explained_variance[:n_modes]
        self._explained_variance_ratio = self._explained_variance_ratio[:n_modes]
        self._pcs = self._pcs[:, :n_modes]

        # Spatial projections
        self._eofs = (self._pcs.conjugate().T @ self.X).T

        # Ensure consistent signs for deterministic output
        maxidx = [abs(self._eofs).argmax(axis=0)]
        flip_signs = np.sign(self._eofs[maxidx, range(self._eofs.shape[1])])
        self._eofs *= flip_signs
        self._pcs *= flip_signs

    def explained_variance(self) -> np.ndarray:
        '''Get the explained variance.

        The explained variance is given by the individual eigenvalues
        of the covariance matrix.

        '''
        return self._explained_variance

    def explained_variance_ratio(self) -> np.ndarray:
        '''Get the explained variance ratio.

        The explained variance ratio is the fraction of total variance
        explained by a given mode and is calculated by :math:`\lambda_i / \sum_i^m \lambda_i`
        where `m` is the total number of modes.

        '''

        return self._explained_variance_ratio

    def eofs(self) -> np.ndarray:
        '''Get the EOFs.

        '''
        return self._eofs

    def pcs(self) -> np.ndarray:
        '''Get the PCs.

        The principal components (PCs), also known as PC scores, are computed
        by projecting the data matrix `X` onto the EOFs.

        '''
        return self._pcs

    def eofs_amplitude(self):
        '''Get the EOF amplitude functions.

        The amplitude is defined as

        .. math::
            A = (V \\oplus \\bar{V})^{1/2}

        where :math:`V` denotes the complex EOFs, :math:`\\bar{V}` the complex
        conjugate and :math:`\\oplus` denotes the Hadamard (element-wise)
        product.

        '''
        return np.sqrt(self._eofs.conj() * self._eofs).real

    def pcs_amplitude(self):
        '''Get the PC amplitude functions.

        The amplitude is defined as

        .. math::
            A = (U \\oplus \\bar{U})^{1/2}

        where :math:`U` denotes the complex PCs, :math:`\\bar{U}` the complex
        conjugate and :math:`\\oplus` denotes the Hadamard (element-wise)
        product.

        '''
        return np.sqrt(self._pcs.conj() * self._pcs).real

    def eofs_phase(self):
        '''Get the EOF phase functions.

        The phase is defined as

        .. math::
            P = \\arctan(I(V) \\oslash R(V))

        where :math:`V` denotes the complex EOFs and :math:`R(\\cdot)`,
        :math:`I(\\cdot)` the real and imaginary part, respectively.
        The :math:`\\oslash` operator represents element-wise division.

        '''
        return np.arctan2(self._eofs.imag, self._eofs.real)

    def pcs_phase(self):
        '''Get the PC phase functions.

        The phase is defined as

        .. math::
            P = \\arctan(I(U) \\oslash R(U))

        where :math:`U` denotes the complex PCs and :math:`R(\\cdot)`,
        :math:`I(\\cdot)` the real and imaginary part, respectively.
        The :math:`\\oslash` operator represents element-wise division.

        '''
        return np.arctan2(self._pcs.imag, self._pcs.real)



from typing import Optional

import xarray as xr

from ..models._base_bootstrapper import _BaseBootstrapper
from ..utils.tools import squeeze


class Bootstrapper(_BaseBootstrapper):
    '''Bootstrapping an xarray EOF model ``xe.xarray.EOF``.

    '''
    def __init__(
            self, n_boot: int,
            alpha : float = 0.05,
            test_type : Optional[str] = 'one-sided'
    ):
        super().__init__(n_boot=n_boot, alpha=alpha, test_type=test_type)

    def bootstrap(self, model):
        super().bootstrap(model)

    def get_params(self):
        return super().get_params()

    def n_significant_modes(self):
        return super().n_significant_modes()

    def explained_variance(self):
        expvar, expvar_mask = super().explained_variance()
        expvar = xr.DataArray(
            expvar,
            dims=['quantile', 'mode'],
            coords=dict(
                quantile=self._params['quantiles'],
                mode=self._model._idx_mode
            ),
            name='explained_variance_uncertainty'
        )
        expvar_mask = xr.DataArray(
            expvar_mask,
            dims=['mode'],
            coords=dict(
                mode=self._model._idx_mode[:-1]
            ),
            name='mode_is_significant'
        )
        return expvar, expvar_mask

    def eofs(self):
        eofs, eofs_mask = super().eofs()
        # Merge quantiles into one DataArray
        qlow, qup = [self._model._tf.back_transform_eofs(eof) for eof in eofs]
        eofs = [
            xr.concat([q0, q1], dim='quantile').assign_coords(dict(quantile=self._params['quantiles']))
            for q0, q1 in zip(qlow, qup)
        ]
        eofs_mask = self._model._tf.back_transform_eofs(eofs_mask)

        # Floats are returned, convert to boolean mask
        eofs_mask = [eof.astype(bool) for eof in eofs_mask]
        eofs_mask = squeeze(eofs_mask)

        for i in range(len(eofs)):
            eofs[i].name = 'eofs_uncertainty'
            eofs_mask[i].name = 'eof_is_significant'

        return squeeze(eofs), squeeze(eofs_mask)

    def pcs(self):
        pcs, pcs_mask = super().pcs()
        # Merge quantiles into one DataArray
        qlow, qup = [self._model._tf.back_transform_pcs(pc) for pc in pcs]
        pcs = xr.concat([qlow, qup], dim='quantile').assign_coords(dict(quantile=self._params['quantiles']))
        pcs_mask = self._model._tf.back_transform_pcs(pcs_mask)

        # Floats are returned, convert to boolean mask
        pcs_mask = pcs_mask.astype(bool)

        pcs.name = 'pcs_uncertainty'
        pcs_mask.name = 'pc_is_significant'

        return squeeze(pcs), squeeze(pcs_mask)



import numpy as np
import scipy as sc
from typing import Optional, Union, List, Tuple

from ..utils.rotation import promax
from ..utils.tools import get_mode_selector


class _BaseMCARotator:
    '''Rotates a solution obtained from ``xe.models.MCA``.'''

    def __init__(
        self,
        n_rot : int,
        power : int = 1,
        loadings : str = 'standard',
        max_iter : int = 1000,
        rtol : float = 1e-8
    ) -> None:
        '''
        Parameters
        ----------
        model : xe.models.EOF
            A EOF model solution.
        n_rot : int
            Number of modes to be rotated.
        loadings : ['standard', 'squared']
            Rotate the singular vectors either scaled by the square root of
            singular values ('standard') or by the singular values ('squared').
            Using 'standard' loadings, the covariance of the rotated modes will be
            conserved i.e. the sum of the covariance of all rotated modes will
            be equal to the unrotated modes. Using 'squared' loadings, the squared
            covariance will be conserved which then can be used to estimate
            the modes' importance.
        power : int
            Defines the power of Promax rotation. Choosing ``power=1`` equals
            a Varimax solution (the default is 1).
        max_iter : int
            Number of maximal iterations for obtaining the rotation matrix
            (the default is 1000).
        rtol : float
            Relative tolerance to be achieved for early stopping the iteration
            process (the default is 1e-8).

        '''
        self._params = {
            'n_rot': n_rot,
            'power': power,
            'loadings': loadings,
            'max_iter': max_iter,
            'rtol': rtol,
        }
        valid_loadingss = ['standard', 'squared']
        if loadings not in valid_loadingss:
            msg = '{:} is not a valid optione. Muste be one of {:}'
            msg = msg.format(loadings, valid_loadingss)
            raise ValueError(msg)

    def rotate(self, model) -> None:
        n_rot = self._params['n_rot']
        power = self._params['power']
        max_iter = self._params['max_iter']
        rtol = self._params['rtol']
        loadings = self._params['loadings']

        self._model = model

        # Construct combined vector of loadingss
        # NOTE: Dunkerton & Cheng (DC) "load" the combined vectors
        # with the square root of the singular values as it is done for
        # standard Varimax rotation. In this case, the total amount of
        # covariance is conserved under rotation. However, for MCA one
        # typically looks at the squared covariance to investigate the
        # importance of a given mode. Using the approach of DC, the squared
        # covariance is not conserved under rotation thus one cannot estimate
        # the modes' importance after rotation. One way around this issue
        # is to rotate the singular vectors loaded with the singular values
        # ("squared loadingss" as opposed to the square root of the singular
        # values. Like this, the squared covariance is conserved and the
        # importance can be estimated.
        if loadings == 'standard':
            # Original implementation by Dunkerton & Cheng;
            # covariance is conserved, squared covariance is not conserved
            scaling = np.sqrt(model._singular_values[:n_rot])
        elif loadings == 'squared':
            scaling = model._singular_values[:n_rot]

        L = np.concatenate(
            [model._Vx[:, :n_rot], model._Vy[:, :n_rot]],
            axis=0
        )
        L *= scaling

        # Rotate loadingss
        L_rot, rot_mat, corr_mat = promax(
            L, power=power, max_iter=max_iter, rtol=rtol
        )
        # Store rotation and correlation matrix
        self._rot_mat = rot_mat
        self._corr_mat = corr_mat

        # Rotated (loaded) singular vectors for X and Y
        Lx_rot = L_rot[:self._model.n_features_X]
        Ly_rot = L_rot[self._model.n_features_X:]

        # Normalization factor of singular vectors
        self._norm_X = np.linalg.norm(Lx_rot, axis=0)
        self._norm_Y = np.linalg.norm(Ly_rot, axis=0)

        # Rotated (normalized) singular vectors
        self._Vx = Lx_rot / self._norm_X
        self._Vy = Ly_rot / self._norm_Y

        # Remove the squaring introduced from scaling
        if loadings == 'squared':
            self._norm_X = np.sqrt(self._norm_X)
            self._norm_Y = np.sqrt(self._norm_Y)

        # Reorder according to covariance
        self._singular_values = self._norm_X * self._norm_Y
        self._idx_var = np.argsort(self._singular_values)[::-1]

        self._norm_X = self._norm_X[self._idx_var]
        self._norm_Y = self._norm_Y[self._idx_var]
        self._singular_values = self._singular_values[self._idx_var]

        self._Vx = self._Vx[:, self._idx_var]
        self._Vy = self._Vy[:, self._idx_var]

        # Explained covariance and fraction
        self._explained_covariance = self._singular_values
        self._squared_covariance_fraction = np.power(self._singular_values, 2) / self._model._squared_total_variance

        # Rotate PCs using rotatation matrix
        self._Ux = self._model._Ux[:, :n_rot]
        self._Uy = self._model._Uy[:, :n_rot]
        R = self._rotation_matrix(inverse_transpose=True)
        self._Ux = self._Ux @ R
        self._Uy = self._Uy @ R
        # Reorder according to variance
        self._Ux = self._Ux[:, self._idx_var]
        self._Uy = self._Uy[:, self._idx_var]

        # Ensure consistent signs for deterministic output
        maxidx = [abs(L_rot).argmax(axis=0)]
        self._mode_signs = np.sign(L_rot[maxidx, range(L_rot.shape[1])])
        self._Vx *= self._mode_signs
        self._Vy *= self._mode_signs
        self._Ux *= self._mode_signs
        self._Uy *= self._mode_signs

    def _rotation_matrix(self, inverse_transpose : bool = False) -> np.ndarray:
        '''Return the rotation matrix.

        Parameters
        ----------
        inverse_transpose : boolean
            If True, return the inverse transposed of the rotation matrix.
            For orthogonal rotations (Varimax) the inverse transpose equals
            the rotation matrix itself. For oblique rotations (Promax), it
            will be different in general (the default is False).

        Returns
        -------
        rotation_matrix : np.ndarray

        '''
        R = self._rot_mat
        power = self._params['power']

        # only for oblique rotations
        # If rotation is orthogonal: R == R^(-1).T
        # If rotation is oblique (i.e. power>1): R != R^(-1).T
        if inverse_transpose and power > 1:
            R = np.linalg.pinv(R).conjugate().T

        return R

    def singular_values(self) -> np.ndarray:
        '''Get the singular values of the covariance matrix.

        '''
        return self._singular_values

    def explained_covariance(self) -> np.ndarray:
        '''Get the explained covariance.

        The explained covariance is simply given by the individual singular
        values of the covariance matrix.

        '''
        return self._explained_covariance

    def squared_covariance_fraction(self) -> np.ndarray:
        '''Get the squared covariance fraction.

        The squared covariance fraction (SCF) is the fraction of total covariance
        explained by a given mode `i` and is calculated by

        .. math::
           SCF_i = \\frac{\sigma_i^2}{\sum_{i=1}^{m} \sigma_i^2}

        where `m` is the total number of modes and :math:`\sigma_i` denotes the
        `ith` singular value of the covariance matrix.

        '''
        return self._squared_covariance_fraction

    def singular_vectors(
            self, scaling : int = 0
    ) -> Tuple[np.ndarray, np.ndarray]:
        '''Get the singular vectors of the covariance matrix.

        Parameters
        ----------
        scaling : [0, 1, 2]
            Singular vectors are scaled (i) to be orthonormal (``scaling=0``),
            (ii) by the square root of the singular values (``scaling=1``) or
            (iii) by the singular values (``scaling=2``) (the default is 0).

        '''
        if scaling == 0:
            Vx = self._Vx
            Vy = self._Vy
        elif scaling == 1:
            Vx = self._Vx * self._norm_X
            Vy = self._Vy * self._norm_Y
        elif scaling == 2:
            Vx = self._Vx * self._norm_X**2
            Vy = self._Vy * self._norm_Y**2
        else:
            msg = (
                'The scaling option {:} is not valid. Please choose one '
                'of the following: [0, 1, 2]'
            )
            msg = msg.format(scaling)
            raise ValueError(msg)
        return Vx, Vy

    def pcs(self, scaling : int = 0) -> Tuple[np.ndarray, np.ndarray]:
        '''Get the PCs.

        The principal components (PCs) in MCA are defined as the projection of
        the data matrices ``X`` and ``Y`` onto the singular vectors.
        There is one set of PCs for each data field.

        Parameters
        ----------
        scaling : [0, 1, 2]
            PCs are scaled (i) to be orthonormal (``scaling=0``), (ii) by the
            square root of the singular values (``scaling=1``) or (iii) by the
            singular values (``scaling=2``) (the default is 0).

        '''
        if scaling == 0:
            Ux = self._Ux
            Uy = self._Uy
        elif scaling == 1:
            Ux = self._Ux * self._norm_X
            Uy = self._Uy * self._norm_Y
        elif scaling == 2:
            Ux = self._Ux * self._norm_X**2
            Uy = self._Uy * self._norm_Y**2
        else:
            msg = (
                'The scaling option {:} is not valid. Please choose one '
                'of the following: [0, 1, 2]'
            )
            msg = msg.format(scaling)
            raise ValueError(msg)
        return Ux, Uy

    def _corr_coef(self, A, B):
        # Columnwise mean of input arrays & subtract from input arrays
        # themeselves
        A_mA = A - A.mean(0)
        B_mB = B - B.mean(0)

        # Sum of squares across rows
        ssA = (A_mA**2).sum(0)
        ssB = (B_mB**2).sum(0)

        # Finally get corr coeff
        return np.dot(A_mA.T, B_mB) / np.sqrt(np.dot(ssA[:, None], ssB[None]))

    def homogeneous_patterns(self) -> Tuple[np.ndarray, np.ndarray]:
        '''Correlation coefficients between PCs and their associated input data.

        More precisely, the homogeneous patterns `r_{hom}` are defined as

        .. math::
          r_{hom, x} = \\corr \\left(X, PC_x \\right)
        .. math::
          r_{hom, y} = \\corr \\left(Y, PC_y \\right)

        Returns
        -------
        Tuple[np.ndarray, np.ndarry]
            Homogeneous patterns for ``X`` and ``Y`` (n_features x n_modes).

        Tuple[np.ndarray, np.ndarry]
             coefficients and associated
            Two-sided p-values for ``X and ``Y`` (n_features x n_modes).
        '''

        # Compute correlation matrix
        corr_X = self._corr_coef(self._model.X, self._Ux)
        corr_Y = self._corr_coef(self._model.Y, self._Uy)

        # Compute two-sided p-values
        # Reference: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.pearsonr.html#r8c6348c62346-1
        a = self._model.n_samples / 2 - 1
        dist = sc.stats.beta(a, a, loc=-1, scale=2)
        pvals_X = 2 * dist.cdf(-abs(corr_X))
        pvals_Y = 2 * dist.cdf(-abs(corr_Y))
        return (corr_X, corr_Y), (pvals_X, pvals_Y)

    def heterogeneous_patterns(self) -> Tuple[np.ndarray, np.ndarray]:
        '''Correlation coefficients between PCs and the opposite input data.

        More precisely, the heterogenous patterns `r_{het}` are defined as

        .. math::
          r_{het, x} = \\corr \\left(X, PC_y \\right)
        .. math::
          r_{het, y} = \\corr \\left(Y, PC_x \\right)

        Returns
        -------
        Tuple[np.ndarray, np.ndarry]
            Heterogenous patterns for ``X`` and ``Y`` (n_features x n_modes).

        Tuple[np.ndarray, np.ndarry]
             coefficients and associated
            Two-sided p-values for ``X and ``Y`` (n_features x n_modes).
        '''

        # Compute correlation matrix
        corr_X = self._corr_coef(self._model.X, self._Uy)
        corr_Y = self._corr_coef(self._model.Y, self._Ux)

        # Compute two-sided p-values
        # Reference: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.pearsonr.html#r8c6348c62346-1
        a = self._model.n_samples / 2 - 1
        dist = sc.stats.beta(a, a, loc=-1, scale=2)
        pvals_X = 2 * dist.cdf(-abs(corr_X))
        pvals_Y = 2 * dist.cdf(-abs(corr_Y))
        return (corr_X, corr_Y), (pvals_X, pvals_Y)

    def reconstruct_XY(
        self,
        mode : Optional[Union[int, List[int], slice]] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        '''Reconstruct original data fields ``X`` and ``Y``.

        If weights were applied, ``X`` and ``Y`` will be automatically
        rescaled.

        Parameters
        ----------
        mode : Optional[Union[int, List[int], slice]]
            Mode(s) based on which ``X`` will be reconstructed. If ``mode`` is
            an int, a single mode is used. If a list of integers is provided,
            use all specified modes for reconstruction. Alternatively, you may
            want to select a slice to reconstruct. The first mode is denoted
            by 1 (and not by 0). If None then ``X`` is recontructed using all
            available modes (the default is None).

        Examples
        --------

        Perform an analysis using some data ``X``:

        >>> model = MCA(X, Y, norm=True)
        >>> model.solve()

        Reconstruct ``X`` and ``Y`` using all modes:

        >>> model.reconstruct_XY()

        Reconstruct ``X`` and ``Y`` using the first mode only:

        >>> model.reconstruct_XY(1)

        Reconstruct ``X`` and ``Y`` using mode 1, 3 and 4:

        >>> model.reconstruct_XY([1, 3, 4])

        Reconstruct ``X`` and ``Y`` using all modes up to mode 10 (including):

        >>> model.reconstruct_XY(slice(10))

        Reconstruct ``X`` and ``Y`` using every second mode between 4 and 8 (both
        including):

        >>> model.reconstruct_XY(slice(4, 8, 2))


        '''
        # Select modes to reconstruct X
        mode = get_mode_selector(mode)

        Vx = self._Vx[:, mode]
        Vy = self._Vy[:, mode]

        Ux = self._Ux[:, mode] * self._norm_X[mode]
        Uy = self._Uy[:, mode] * self._norm_Y[mode]

        Xrec = Ux @ Vx.conj().T
        Yrec = Uy @ Vy.conj().T
        # Unweight and add mean
        Xrec = (Xrec / self._model._weights_X) + self._model._X_mean
        Yrec = (Yrec / self._model._weights_Y) + self._model._Y_mean
        return Xrec, Yrec

    def project_onto_left_singular_vectors(
        self,
        X : np.ndarray = None,
        scaling : int = 0
    ) -> np.ndarray:
        '''Project new data onto the singular vectors.

        Parameters
        ----------
        X : np.ndarray
             New data to project onto first singular vector.
        scaling : [0, 1, 2]
            Projections are scaled (i) to be orthonormal (``scaling=0``), (ii) by the
            square root of the singular values (``scaling=1``) or (iii) by the
            singular values (``scaling=2``) (the default is 0).

        Returns
        -------
        np.ndarray
            Projections of new data onto left singular vector

        '''
        X = X.copy()
        # Remove mean and apply weights
        try:
            X -= self._model._X_mean
        except ValueError:
            err_msg = (
                'New data has invalid feature dimensions and cannot be '
                'projected onto singular vectors. Expected shapes'
            )
            raise ValueError(err_msg)
        X *= self._model._weights_X

        # Project new data onto singular values
        n_rot = self._params['n_rot']
        Ux = X @ self._model._Vx[:, :n_rot] / np.sqrt(self._singular_values)
        # Rotate
        R = self._rotation_matrix(inverse_transpose=True)
        Ux = Ux @ R
        # Reorder
        Ux = Ux[:, self._idx_var]
        # Flip sign
        Ux = Ux * self._mode_signs

        # Apply scaling
        if scaling == 0:
            pass
        elif scaling == 1:
            Ux *= self._norm_X
        elif scaling == 2:
            Ux *= self._norm_X**2
        else:
            err_msg = (
                'Scaling option {:} is not valid but must be one '
                'of [0, 1, 2]'
            )
            err_msg = err_msg.foramt(scaling)
            raise ValueError(err_msg)
        return Ux

    def project_onto_right_singular_vectors(
        self,
        Y : np.ndarray = None,
        scaling : int = 0
    ) -> np.ndarray:
        '''Project new data onto the singular vectors.

        Parameters
        ----------
        Y : np.ndarray
             New data to project onto second singular vector.
        scaling : [0, 1, 2]
            Projections are scaled (i) to be orthonormal (``scaling=0``), (ii) by the
            square root of the singular values (``scaling=1``) or (iii) by the
            singular values (``scaling=2``) (the default is 0).

        Returns
        -------
        np.ndarray
            Projections of new data onto left singular vector

        '''
        Y = Y.copy()
        # Remove mean and apply weights
        try:
            Y -= self._model._Y_mean
        except ValueError:
            err_msg = (
                'New data has invalid feature dimensions and cannot be '
                'projected onto singular vectors. Expected shapes'
            )
            raise ValueError(err_msg)
        Y *= self._model._weights_Y

        # Project new data onto singular values
        n_rot = self._params['n_rot']
        Uy = Y @ self._model._Vy[:, :n_rot] / np.sqrt(self._singular_values)
        # Rotate
        R = self._rotation_matrix(inverse_transpose=True)
        Uy = Uy @ R
        # Reorder
        Uy = Uy[:, self._idx_var]
        # Flip sign
        Uy = Uy * self._mode_signs

        # Apply scaling
        if scaling == 0:
            pass
        elif scaling == 1:
            Uy *= self._norm_Y
        elif scaling == 2:
            Uy *= self._norm_Y**2
        else:
            err_msg = (
                'Scaling option {:} is not valid but must be one '
                'of [0, 1, 2]'
            )
            err_msg = err_msg.foramt(scaling)
            raise ValueError(err_msg)
        return Uy



from typing import Optional, Union, Iterable, Tuple, List

import numpy as np

from xeofs.models._base_eof import _BaseEOF
from xeofs.models._transformer import _MultiArrayTransformer
from ..utils.tools import squeeze

import warnings

warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

Array = np.ndarray
ArrayList = Union[Array, List[Array]]


class EOF(_BaseEOF):
    '''EOF analysis of a single or multiple ``np.ndarray``.

    Parameters
    ----------
    X : np.ndarray
        Data matrix to be decomposed. ``X`` can be any N-dimensional array.
        Dimensions whose variance shall be maximised (denoted as *samples*)
        have to be defined by the ``axis`` parameter. All remaining axes will
        be reshaped into a new axis called *features*.
    n_modes : Optional[int]
        Number of modes to compute. Computing less modes can results in
        performance gains. If None, then the maximum number of modes is
        equivalent to ``min(n_samples, n_features)`` (the default is None).
    norm : bool
        Normalize each feature (e.g. grid cell) by its temporal standard
        deviation (the default is False).
    weights : Optional[np.ndarray] = None
        Weights applied to features. Must have the same dimensions as the
        original features which are the remaining axes not specified by
        ``axis`` parameter).
    axis : Union[int, Iterable[int]]
        Axis along which variance should be maximised. Can also be
        multi-dimensional. For example, given a data array of dimensions
        ``(n x p1 x p2)`` with `n` time series at ``p1`` and ``p2`` different
        locations, ``axis=0`` will maximise `temporal` variance along ``n``.
        In contrast, ``axis=[1, 2]`` will maximise `spatial` variance along
        ``(p1 x p2)`` (the default is 0).


    Examples
    --------

    Import package and create data:

    >>> import numpy as np
    >>> from xeofs.models import EOF
    >>> rng = np.random.default_rng(7)
    >>> X = rng.standard_normal((4, 2, 3))

    Initialize standardized EOF analysis and compute the first 2 modes:

    >>> model = EOF(X, norm=True, n_modes=2)
    >>> model.solve()

    Get explained variance:

    >>> model.explained_variance()
    ... array([4.81848833, 2.20765019])

    Get EOFs:

    >>> model.eofs()
    ... array([[[ 0.52002146, -0.00698575],
    ...         [ 0.41059796, -0.31603563]],
    ...
    ...         [[ 0.32091783,  0.47647144],
    ...         [-0.51771611,  0.01380736]],
    ...
    ...         [[-0.28840959,  0.63341346],
    ...         [ 0.32678537,  0.52119516]]])

    Get PCs:

    >>> model.pcs()
    ... array([[ 1.76084189, -0.92030205],
    ...        [-2.13581896, -1.49704775],
    ...        [ 2.02091078,  0.65502667],
    ...        [-1.64593371,  1.76232312]])

    '''

    def __init__(
        self,
        X: ArrayList,
        axis : Union[int, Iterable[int]] = 0,
        n_modes : Optional[int] = None,
        norm : bool = False,
        weights : Optional[ArrayList] = None,
    ):

        self._tf = _MultiArrayTransformer()
        X = self._tf.fit_transform(X, axis=axis)
        weights = self._tf.transform_weights(weights)

        super().__init__(
            X=X,
            n_modes=n_modes,
            norm=norm,
            weights=weights
        )

    def eofs(self, scaling : int = 0) -> ArrayList:
        eofs = super().eofs(scaling=scaling)
        eofs = self._tf.back_transform_eofs(eofs)
        return squeeze(eofs)

    def pcs(self, scaling : int = 0) -> Array:
        pcs = super().pcs(scaling=scaling)
        return self._tf.back_transform_pcs(pcs)

    def eofs_as_correlation(self) -> Tuple[ArrayList, ArrayList]:
        corr, pvals = super().eofs_as_correlation()
        corr = self._tf.back_transform_eofs(corr)
        pvals = self._tf.back_transform_eofs(pvals)
        return squeeze(corr), squeeze(pvals)

    def reconstruct_X(
        self,
        mode : Optional[Union[int, List[int], slice]] = None
    ) -> ArrayList:
        Xrec = super().reconstruct_X(mode)
        Xrec = self._tf.back_transform(Xrec)
        return squeeze(Xrec)

    def project_onto_eofs(
        self,
        X : ArrayList,
        scaling : int = 0
    ) -> Array:
        '''Project new data onto the EOFs.

        Parameters
        ----------
        X : array or list of arrays
             New data to project. Data must have same feature shape as original
             data.
        scaling : [0, 1, 2]
            Projections are scaled (i) to be orthonormal (``scaling=0``), (ii) by the
            square root of the eigenvalues (``scaling=1``) or (iii) by the
            singular values (``scaling=2``). In case no weights were applied,
            scaling by the singular values results in the projections having the
            unit of the input data (the default is 0).

        '''
        proj = _MultiArrayTransformer()
        X_proj = proj.fit_transform(X, axis=self._tf.axis_samples)
        pcs = super().project_onto_eofs(X=X_proj, scaling=scaling)
        return proj.back_transform_pcs(pcs)



from typing import Optional

import numpy as np
import pandas as pd

from ..models._base_bootstrapper import _BaseBootstrapper
from .eof import EOF
from ..utils.tools import squeeze


class Bootstrapper(_BaseBootstrapper):
    '''Bootstrapping a pandas EOF model ``xe.pandas.EOF``.

    '''
    def __init__(
            self, n_boot : int,
            alpha : float = 0.05,
            test_type : Optional[str] = 'one-sided'
    ):
        super().__init__(n_boot=n_boot, alpha=alpha, test_type=test_type)

    def bootstrap(self, model : EOF):
        super().bootstrap(model)

    def get_params(self):
        return super().get_params()

    def n_significant_modes(self):
        return super().n_significant_modes()

    def explained_variance(self):
        expvar, expvar_mask = super().explained_variance()
        expvar = pd.DataFrame(
            expvar.T,
            columns=pd.Index(self._params['quantiles'], name='quantile'),
            index=self._model._idx_mode,
        )
        expvar_mask = pd.DataFrame(
            expvar_mask,
            columns=['is_significant'],
            index=self._model._idx_mode[:-1]
        )

        return expvar, expvar_mask

    def eofs(self):
        eofs, eofs_mask = super().eofs()
        eofs = [squeeze(self._model._tf.back_transform_eofs(e)) for e in eofs]
        eofs_mask = squeeze(self._model._tf.back_transform_eofs(eofs_mask))

        return eofs, eofs_mask

    def pcs(self):
        pcs, pcs_mask = super().pcs()
        pcs = [squeeze(self._model._tf.back_transform_pcs(q)) for q in pcs]
        pcs_mask = squeeze(self._model._tf.back_transform_pcs(pcs_mask))
        return pcs, pcs_mask



''' Implementation of VARIMAX and PROMAX rotation. '''

# =============================================================================
# Imports
# =============================================================================
import numpy as np


# =============================================================================
# VARIMAX
# =============================================================================
def varimax(
    X : np.ndarray,
    gamma : float = 1,
    max_iter : int = 1000,
    rtol : float = 1e-8
):
    '''
    Perform (orthogonal) Varimax rotation.

    This implementation also works for complex numbers.

    Parameters
    ----------
    X : np.ndarray
        2D matrix to be rotated containing features as rows and modes as
        columns.
    gamma : float
        Parameter which determines the type of rotation performed: varimax (1),
        quartimax (0). Other values are possible. The default is 1.
    max_iter : int
        Number of iterations performed. The default is 1000.
    rtol : float
        Relative tolerance at which iteration process terminates.
        The default is 1e-8.

    Returns
    -------
    Xrot : np.ndarray
        Rotated matrix with same dimensions as X.
    R : array-like
        Rotation matrix of shape ``(n_rot x n_rot)``

    '''
    X = X.copy()
    n_samples, n_modes = X.shape

    if n_modes < 2:
        err_msg = 'Cannot rotate {:} modes (columns), but must be 2 or more.'
        err_msg = err_msg.format(n_modes)
        raise ValueError(err_msg)

    # Initialize rotation matrix
    R = np.eye(n_modes)

    # Normalize the matrix using square root of the sum of squares (Kaiser)
    h = np.sqrt(np.sum(X * X.conjugate(), axis=1))
    # A = np.diag(1./h) @ A

    # Add a stabilizer to avoid zero communalities
    eps = 1e-9
    X = (1. / (h + eps))[:, np.newaxis] * X

    # Seek for rotation matrix based on varimax criteria
    delta = 0.
    converged = False
    for i in range(max_iter):
        delta_old = delta
        basis = X @ R

        basis2 = basis * basis.conjugate()
        basis3 = basis2 * basis
        W = np.diag(np.sum(basis2, axis=0))
        alpha = gamma / n_samples

        transformed = X.conjugate().T @ (basis3 - (alpha * basis @ W))
        U, svals, VT = np.linalg.svd(transformed)
        R = U @ VT
        delta = np.sum(svals)
        if (abs(delta - delta_old) / delta) < rtol:
            converged = True
            break

    if(not converged):
        raise RuntimeError('Rotation process did not converge.')

    # De-normalize
    X = h[:, np.newaxis] * X

    # Rotate
    Xrot = X @ R
    return Xrot, R


# =============================================================================
# PROMAX
# =============================================================================
def promax(
    X : np.ndarray,
    power : int = 1,
    max_iter : int = 1000,
    rtol : float = 1e-8
):
    '''
    Perform (oblique) Promax rotation.

    This implementation also works for complex numbers.

    Parameters
    ----------
    X : np.ndarray
        2D matrix to be rotated. Must have shape ``p x m`` containing
        p features and m modes.
    power : int
        Rotation parameter defining the power the Varimax solution is raised
        to. For ``power=1``, this is equivalent to the Varimax solution
        (the default is 1).
    max_iter: int
        Maximum number of iterations for finding the rotation matrix
        (the default is 1000).
    rtol:
        The relative tolerance for the rotation process to achieve
        (the default is 1e-8).

    Returns
    -------
    Xrot : np.ndarray
        2D matrix containing the rotated modes.
    rot_mat : np.ndarray
        Rotation matrix of shape ``m x m`` with m being number of modes.
    phi : np.ndarray
        Correlation matrix of PCs of shape ``m x m`` with m being number
        of modes. For Varimax solution (``power=1``), the correlation matrix
        is diagonal i.e. the modes are uncorrelated.

    '''
    X = X.copy()

    # Perform varimax rotation
    X, rot_mat = varimax(X=X, max_iter=max_iter, rtol=rtol)

    # Pre-normalization by communalities (sum of squared rows)
    h = np.sqrt(np.sum(X * X.conjugate(), axis=1))
    # Add a stabilizer to avoid zero communalities
    eps = 1e-9
    X = (1. / (h + eps))[:, np.newaxis] * X

    # Max-normalisation of columns
    Xnorm = X / np.max(abs(X), axis=0)

    # "Procustes" equation
    P = Xnorm * np.abs(Xnorm)**(power - 1)

    # Fit linear regression model of "Procrustes" equation
    # see Richman 1986 for derivation
    L = np.linalg.inv(X.conjugate().T @ X) @ X.conjugate().T @ P

    # calculate diagonal of inverse square
    try:
        sigma_inv = np.diag(np.diag(np.linalg.inv(L.conjugate().T @ L)))
    except np.linalg.LinAlgError:
        sigma_inv = np.diag(np.diag(np.linalg.pinv(L.conjugate().T @ L)))

    # transform and calculate inner products
    L = L @ np.sqrt(sigma_inv)
    Xrot = X @ L

    # Post-normalization based on Kaiser
    Xrot = h[:, np.newaxis] * Xrot

    rot_mat = rot_mat @ L

    # Correlation matrix
    L_inv = np.linalg.inv(L)
    phi = L_inv @ L_inv.conjugate().T

    return Xrot, rot_mat, phi



from typing import Optional, Tuple, List, Union, Iterable

import xarray as xr

from ..models._base_rock_pca import _BaseROCK_PCA
from ..utils.tools import squeeze
from ._transformer import _MultiDataArrayTransformer

DataArray = xr.DataArray
DataArrayList = Union[DataArray, List[DataArray]]


class ROCK_PCA(_BaseROCK_PCA):
    '''ROCK-PCA of a single or multiple ``xr.DataArray``.'''

    def __init__(
        self,
        X: DataArray,
        n_rot : int,
        power : int,
        sigma : float,
        dim: Union[str, Iterable[str]] = 'time',
        n_modes: Optional[int] = None,
        norm: bool = False,
        weights: Optional[DataArray] = None,
    ):
        use_coslat = True if weights == 'coslat' else False

        self._tf = _MultiDataArrayTransformer()
        X = self._tf.fit_transform(X, dim=dim, coslat=use_coslat)

        if use_coslat:
            weights = self._tf.coslat_weights
        weights = self._tf.transform_weights(weights)

        super().__init__(
            X=X, n_rot=n_rot, power=power, sigma=sigma, n_modes=n_modes,
            norm=norm, weights=weights
        )
        self._idx_mode = xr.IndexVariable('mode', range(1, self._params['n_modes'] + 1))
        self._dim = dim

    def explained_variance(self) -> DataArray:
        expvar = super().explained_variance()
        return xr.DataArray(
            expvar,
            dims=['mode'],
            coords={'mode' : self._idx_mode},
            name='explained_variance'
        )

    def explained_variance_ratio(self) -> DataArray:
        expvar = super().explained_variance_ratio()
        return xr.DataArray(
            expvar,
            dims=['mode'],
            coords={'mode' : self._idx_mode},
            name='explained_variance_ratio'
        )

    def eofs(self) -> DataArrayList:
        eofs = super().eofs()
        eofs = self._tf.back_transform_eofs(eofs)
        return squeeze(eofs)

    def pcs(self) -> DataArray:
        pcs = super().pcs()
        return self._tf.back_transform_pcs(pcs)

    def eofs_amplitude(self) -> DataArrayList:
        amp = super().eofs_amplitude()
        amp = self._tf.back_transform_eofs(amp)
        for da in amp:
            da.name = 'eofs_amplitude'
        return squeeze(amp)

    def pcs_amplitude(self) -> DataArray:
        amp = super().pcs_amplitude()
        amp = self._tf.back_transform_pcs(amp)
        amp.name = 'pcs_amplitude'
        return amp

    def eofs_phase(self) -> DataArrayList:
        phase = super().eofs_phase()
        phase = self._tf.back_transform_eofs(phase)
        for da in phase:
            da.name = 'eofs_phase'
        return squeeze(phase)

    def pcs_phase(self) -> DataArray:
        phase = super().pcs_phase()
        phase = self._tf.back_transform_pcs(phase)
        phase.name = 'pcs_phase'
        return phase



from typing import Optional, Union, Iterable, Tuple, List

import numpy as np

from xeofs.models._base_mca import _BaseMCA
from xeofs.models._transformer import _MultiArrayTransformer
from ..utils.tools import squeeze

import warnings

warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

Array = np.ndarray
ArrayList = Union[Array, List[Array]]


class MCA(_BaseMCA):
    '''Maximum Covariance Analysis (MCA) of a single or multiple ``np.ndarray``.

    MCA is also known as Singular Value Decomposition (SVD) analysis or
    Partial Least Squares (PLS) analysis.

    Parameters
    ----------
    X, Y : np.ndarray
        Data matrices two compute maximum covariance. ``X`` and ``Y`` can be
        any N-dimensional array. Dimensions along which covariance shall be
        maximised (denoted as *samples*) have to be defined by the
        ``axis`` parameter. All remaining axes will be reshaped into a new
        axis called *features*. Sample dimension of ``X`` and ``Y`` must be
        the same.
    n_modes : Optional[int]
        Number of modes to compute. Computing less modes can results in
        performance gains. If None, then the maximum number of modes is
        equivalent to ``min(n_samples, n_features)`` (the default is None).
    norm : bool
        Normalize each feature (e.g. grid cell) by its temporal standard
        deviation (the default is False).
    weights_X : Optional[np.ndarray] = None
        Weights applied to features of ``X``. Must have the same dimensions as the
        original features which are the remaining axes not specified by
        ``axis`` parameter).
    weights_Y : Optional[np.ndarray] = None
        Weights applied to features of ``Y``. Must have the same dimensions as the
        original features which are the remaining axes not specified by
        ``axis`` parameter).
    axis : Union[int, Iterable[int]]
        Axis along which variance should be maximised. Can also be
        multi-dimensional. For example, given a data array of dimensions
        ``(n x p1 x p2)`` with `n` time series at ``p1`` and ``p2`` different
        locations, ``axis=0`` will maximise `temporal` variance along ``n``.
        In contrast, ``axis=[1, 2]`` will maximise `spatial` variance along
        ``(p1 x p2)`` (the default is 0).


    Examples
    --------

    Import package and create data:

    >>> import numpy as np
    >>> from xeofs.models import MCA
    >>> rng = np.random.default_rng(7)
    >>> X = rng.standard_normal((45, 4, 3))
    >>> Y = rng.standard_normal((45, 2, 5))

    Initialize standardized MCA analysis and compute the first 2 modes:

    >>> model = MCA(X, norm=True, n_modes=2, axis=0)
    >>> model.solve()

    Get singular vectors:

    >>> model.singular_vectors()

    Get homogeneous patterns:

    >>> model.homogeneous_patterns()

    '''
    def __init__(
        self,
        X: ArrayList,
        Y: ArrayList,
        axis : Union[int, Iterable[int]] = 0,
        n_modes : Optional[int] = None,
        norm : bool = False,
        weights_X : Optional[ArrayList] = None,
        weights_Y : Optional[ArrayList] = None,
    ):

        self._tfx = _MultiArrayTransformer()
        self._tfy = _MultiArrayTransformer()
        X = self._tfx.fit_transform(X, axis=axis)
        Y = self._tfy.fit_transform(Y, axis=axis)
        weights_X = self._tfx.transform_weights(weights_X)
        weights_Y = self._tfy.transform_weights(weights_Y)

        super().__init__(
            X=X,
            Y=Y,
            n_modes=n_modes,
            norm=norm,
            weights_X=weights_X,
            weights_Y=weights_Y
        )

    def singular_values(self) -> Array:
        return super().singular_values()

    def explained_covariance(self) -> Array:
        return super().explained_covariance()

    def squared_covariance_fraction(self) -> Array:
        return super().squared_covariance_fraction()

    def singular_vectors(
            self, scaling : int = 0
    ) -> Tuple[ArrayList, ArrayList]:
        Vx, Vy = super().singular_vectors(scaling=scaling)
        Vx = self._tfx.back_transform_eofs(Vx)
        Vy = self._tfy.back_transform_eofs(Vy)
        return squeeze(Vx), squeeze(Vy)

    def pcs(self, scaling : int = 0) -> Tuple[ArrayList, ArrayList]:
        Ux, Uy = super().pcs(scaling=scaling)
        Ux = self._tfx.back_transform_pcs(Ux)
        Uy = self._tfy.back_transform_pcs(Uy)
        return Ux, Uy

    def homogeneous_patterns(self) -> Tuple[ArrayList, ArrayList]:
        hom_pats, pvals = super().homogeneous_patterns()
        hom_patsx = squeeze(self._tfx.back_transform_eofs(hom_pats[0]))
        hom_patsy = squeeze(self._tfy.back_transform_eofs(hom_pats[1]))
        pvalsx = squeeze(self._tfx.back_transform_eofs(pvals[0]))
        pvalsy = squeeze(self._tfy.back_transform_eofs(pvals[1]))
        return (hom_patsx, hom_patsy), (pvalsx, pvalsy)

    def heterogeneous_patterns(self) -> Tuple[ArrayList, ArrayList]:
        het_pats, pvals = super().heterogeneous_patterns()
        het_patsx = squeeze(self._tfx.back_transform_eofs(het_pats[0]))
        het_patsy = squeeze(self._tfy.back_transform_eofs(het_pats[1]))
        pvalsx = squeeze(self._tfx.back_transform_eofs(pvals[0]))
        pvalsy = squeeze(self._tfy.back_transform_eofs(pvals[1]))
        return (het_patsx, het_patsy), (pvalsx, pvalsy)

    def reconstruct_XY(
        self,
        mode : Optional[Union[int, List[int], slice]] = None
    ) -> Tuple[ArrayList, ArrayList]:
        Xrec, Yrec = super().reconstruct_XY(mode)
        Xrec = self._tfx.back_transform(Xrec)
        Yrec = self._tfy.back_transform(Yrec)
        return squeeze(Xrec), squeeze(Yrec)

    def project_onto_left_singular_vectors(
        self,
        X : ArrayList = None,
        scaling : int = 0
    ) -> Array:
        # Transform data to 2D
        projx = _MultiArrayTransformer()
        X_proj = projx.fit_transform(X, axis=self._tfx.axis_samples)
        # Perform projection
        pcs_X = super().project_onto_left_singular_vectors(
            X=X_proj, scaling=scaling
        )
        # Backtransform to PC format
        pcs_X = projx.back_transform_pcs(pcs_X)
        return pcs_X

    def project_onto_right_singular_vectors(
        self,
        Y : ArrayList = None,
        scaling : int = 0
    ) -> Array:
        # Transform data to 2D
        projy = _MultiArrayTransformer()
        Y_proj = projy.fit_transform(Y, axis=self._tfy.axis_samples)
        # Perform projection
        pcs_Y = super().project_onto_right_singular_vectors(
            Y=Y_proj, scaling=scaling
        )
        # Backtransform to PC format
        pcs_Y = projy.back_transform_pcs(pcs_Y)
        return pcs_Y



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['xeofs', 'xeofs.models', 'xeofs.pandas', 'xeofs.utils', 'xeofs.xarray']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.2,<2.0.0',
 'pandas>=1.4.1,<2.0.0',
 'pooch>=1.6.0,<2.0.0',
 'scikit-learn>=1.0.2,<2.0.0',
 'tqdm>=4.64.0,<5.0.0',
 'xarray>=0.21.1,<0.22.0']

setup_kwargs = {
    'name': 'xeofs',
    'version': '0.7.0',
    'description': 'Collection of EOF analysis and related techniques for climate science',
    'long_description': '.. image:: examples/1eof/rotated_eof.jpg\n  :align: center\n  :width: 800\n  :alt: Comparison of standard, Varimax-rotated and Proxmax-rotated EOF analysis for temperature field over North America.\n\nExample_ showing sea surface temperature decomposed via EOF analysis, Varimax rotation and Promax rotation.\n\n.. _Example: https://xeofs.readthedocs.io/en/stable/auto_examples/1eof/plot_rotated_eof.html#sphx-glr-auto-examples-1eof-plot-rotated-eof-py\n\n==================================================\nxeofs: Multi-dimensional EOF analysis and variants\n==================================================\n\n|badge_build_status| |badge_docs_status| |badge_version_pypi| |badge_conda_version| |badge_downloads| |badge_coverage| |badge_license| |badge_zenodo|\n\n.. |badge_version_pypi| image:: https://img.shields.io/pypi/v/xeofs\n   :alt: PyPI\n.. |badge_build_status| image:: https://img.shields.io/github/workflow/status/nicrie/xeofs/CI\n   :alt: GitHub Workflow Status (event)\n.. |badge_docs_status| image:: https://readthedocs.org/projects/xeofs/badge/?version=latest\n   :target: https://xeofs.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation Status\n.. |badge_downloads_pypi| image:: https://img.shields.io/pypi/dm/xeofs\n    :alt: PyPI - Downloads\n.. |badge_coverage| image:: https://codecov.io/gh/nicrie/xeofs/branch/main/graph/badge.svg?token=8040ZDH6U7\n    :target: https://codecov.io/gh/nicrie/xeofs\n.. |badge_zenodo| image:: https://zenodo.org/badge/DOI/10.5281/zenodo.6323012.svg\n   :target: https://doi.org/10.5281/zenodo.6323012\n   :alt: DOI - Zenodo\n.. |badge_license| image:: https://img.shields.io/pypi/l/xeofs\n  :alt: License\n.. |badge_conda_version| image:: https://img.shields.io/conda/vn/conda-forge/xeofs\n   :alt: Conda (channel only)\n.. |badge_downloads_conda| image:: https://img.shields.io/conda/dn/conda-forge/xeofs\n   :alt: Conda downloads\n.. |badge_downloads| image:: https://static.pepy.tech/personalized-badge/xeofs?period=total&units=international_system&left_color=grey&right_color=orange&left_text=Downloads\n   :target: https://pepy.tech/project/xeofs\n   :alt: Total downloads\n\nEmpirical orthogonal function (EOF) analysis, more commonly known as\nprincipal component analysis (PCA), is one of the most popular methods\nfor dimension reduction and structure identification in Earth system sciences.\nDue to this popularity, a number of different EOF variants have been developed\nover the last few years, either to mitigate some pitfalls of ordinary EOF\nanalysis (e.g. orthogonality, interpretability, linearity) or to broaden its\nscope (e.g. multivariate variants).\n\nCurrently, there are several implementations of EOF analysis on GitHub that\nfacilitate the acceptance and application of this method by the broader\nscientific community. Each of these implementations has its own strengths,\nwhich need to be highlighted (please `let me know`_, if I forgot any):\n\n\nAvailable Models\n----------------\n\n=====================  ==========  ==========  ==========  ==========  ==========  ==========\nPackage                 **xeofs**   eofs_       pyEOF_      xeof_       xMCA_       xmca2_\n=====================  ==========  ==========  ==========  ==========  ==========  ==========\nEOF analysis           ✅           ✅           ✅           ✅           ✅            ✅\nRotated EOF analysis   ✅           ❌           ✅           ❌           ❌            ✅\nComplex EOF analysis   ❌           ❌           ❌           ❌           ❌            ✅\n`ROCK-PCA`_            ✅           ❌           ❌           ❌           ❌            ❌\nMultivariate EOF       ✅           ✅           ❌           ❌           ❌            ❌\nMCA                    ✅           ❌           ❌           ❌           ✅            ✅\nRotated MCA            ✅           ❌           ❌           ❌           ❌            ✅\nComplex MCA            ❌           ❌           ❌           ❌           ❌            ✅\nMultivariate MCA       ✅           ❌           ❌           ❌           ❌            ❌\n=====================  ==========  ==========  ==========  ==========  ==========  ==========\n\n.. _ROCK-PCA: https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=8989964&casa_token=3zKG0dtp-ewAAAAA:FM1CrVISSSqhWEAwPGpQqCgDYccfLG4N-67xNNDzUBQmMvtIOHuC7T6X-TVQgbDg3aDOpKBksg&tag=1\n\n\nAdditional features\n----------------------\n\n=====================  ==========  ==========  ==========  ==========  ==========  ==========\nPackage                 **xeofs**  eofs_       pyEOF_      xeof_       xMCA_       xmca2_\n=====================  ==========  ==========  ==========  ==========  ==========  ==========\n``numpy`` interface    ✅           ✅           ❌           ❌           ❌           ✅\n``pandas`` interface   ✅           ❌           ❌           ❌           ❌           ❌\n``xarray`` interface   ✅           ?           ✅           ✅           ✅           ✅\nFast algorithm         ✅           ❌           ✅           ❌           ❌           ❌\nDask support           ❌           ✅           ❌           ✅           ❌           ❌\nMulti-dimensional      ✅           ❌           ❌           ❌           ❌           ❌\nSignificance analysis  ✅           ❌           ❌           ❌           ❌           ❌\n=====================  ==========  ==========  ==========  ==========  ==========  ==========\n\n\n.. _eofs: https://github.com/ajdawson/eofs\n.. _xeof: https://github.com/dougiesquire/xeof\n.. _xMCA: https://github.com/Yefee/xMCA\n.. _pyEOF: https://github.com/zzheng93/pyEOF\n.. _xmca2: https://github.com/nicrie/xmca\n\n.. _let me know: niclasrieger@gmail.com\n\n\nWhy ``xeofs``?\n----------------------\n\nThe goal of ``xeofs`` is to merge these different implementations and to simplify the integration of other existing and future variants of EOF analysis thanks to its modular code structure.\nThe official name is deliberately chosen to be similar to the other implementations to make it clear that ``xeofs`` is nothing revolutionary new in itself. The point is not to distinguish this implementation from the others, but rather to unify (+ extend) already existing implementations.\n\nThis project is intended to be a collaborative project of the scientific community and the contribution of EOF variants in the form of pull requests is explicitly encouraged.\nIf you are interested, just `contact me`_ or open an `Issue`_.\n\n.. _contact me: niclasrieger@gmail.com\n.. _Issue: https://github.com/nicrie/xeofs/issues\n\n\n\nInstallation\n----------------------\n\nIf you are using ``conda``, it is recommend to install via:\n\n.. code-block:: ini\n\n  conda install -c conda-forge xeofs\n\nAlternatively, you can install the package through ``pip``:\n\n.. code-block:: ini\n\n  pip install xeofs\n\n\nHow to use it?\n----------------------\nDocumentation_ is work in progress. Meanwhile check out some examples_ to get started:\n\n+ EOF analysis (S-mode_)\n+ EOF analysis (T-mode_)\n+ Rotated_ EOF analysis (Varimax, Promax)\n+ Weighted_ EOF analysis\n+ Multivariate_ EOF analysis\n+ Significance analysis via bootstrapping\n+ Maximum Covariance Analysis\n\n.. _T-mode: https://xeofs.readthedocs.io/en/latest/auto_examples/1eof/plot_eof-tmode.html#sphx-glr-auto-examples-1eof-plot-eof-tmode-py\n.. _S-mode: https://xeofs.readthedocs.io/en/latest/auto_examples/1eof/plot_eof-smode.html#sphx-glr-auto-examples-1eof-plot-eof-smode-py\n.. _Weighted: https://xeofs.readthedocs.io/en/latest/auto_examples/1eof/plot_weighted_eof.html#sphx-glr-auto-examples-1eof-plot-weighted-eof-py\n.. _Rotated: https://xeofs.readthedocs.io/en/latest/auto_examples/1eof/plot_rotated_eof.html#sphx-glr-auto-examples-1eof-plot-rotated-eof-py\n.. _Multivariate: https://xeofs.readthedocs.io/en/latest/auto_examples/1eof/plot_multivariate-eof-analysis.html#sphx-glr-auto-examples-1eof-plot-multivariate-eof-analysis-py\n.. _Documentation: https://xeofs.readthedocs.io/en/latest/\n.. _examples: https://xeofs.readthedocs.io/en/latest/auto_examples/index.html\n\n\n\nCredits\n----------------------\n\n- to Andrew Dawson_ for the first and fundamental Python package for EOF analysis\n- to Yefee_ from which I took some inspiration to implement MCA\n- to James Chapman_ who created a great Python package for Canonical Correlation Analysis\n- to Diego Bueso_ for his open-source ROCK-PCA implementation in Matlab\n- to yngvem_ for how to organize the project folder structure\n- to all the developers of NumPy_, pandas_ \\& xarray_ for their invaluable contributions to science\n\n\n.. _NumPy: https://www.numpy.org\n.. _pandas: https://pandas.pydata.org\n.. _xarray: https://xarray.pydata.org\n.. _yngvem: https://github.com/yngvem/python-project-structure\n.. _pooch: https://github.com/fatiando/pooch\n.. _Chapman: https://github.com/jameschapman19/cca_zoo\n.. _Bueso: https://github.com/DiegoBueso/ROCK-PCA\n.. _Dawson: https://github.com/ajdawson/eofs\n.. _Yefee: https://github.com/Yefee/xMCA\n\n\nHow to cite?\n----------------------\nPlease make sure that when using ``xeofs`` you always cite the **original source** of the method used. Additionally, if you find ``xeofs`` useful for your research, you may cite it as follows::\n\n   @software{rieger_xeofs_2022,\n     title = {xeofs: Multi-dimensional {EOF} analysis and variants in xarray},\n     url = {https://github.com/nicrie/xeofs}\n     version = {0.6.0},\n     author = {Rieger, Niclas},\n     date = {2022},\n     doi = {10.5281/zenodo.6323011}\n   }\n',
    'author': 'Niclas Rieger',
    'author_email': 'niclasrieger@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nicrie/xeofs',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import xarray as xr
from typing import Optional, Union, List, Tuple

from ._transformer import _MultiDataArrayTransformer
from ..models._base_mca_rotator import _BaseMCARotator
from ..utils.tools import squeeze

DataArray = xr.DataArray
DataArrayList = Union[DataArray, List[DataArray]]


class MCA_Rotator(_BaseMCARotator):
    '''Rotates a solution obtained from ``xe.xarray.MCA``.'''

    def __init__(
        self,
        n_rot : int,
        loadings : str = 'standard',
        power : int = 1,
        max_iter : int = 1000,
        rtol : float = 1e-8
    ):

        super().__init__(
            n_rot=n_rot, loadings=loadings, power=power,
            max_iter=max_iter, rtol=rtol
        )

    def singular_values(self) -> DataArray:
        n_rot = self._params['n_rot']
        return xr.DataArray(
            super().singular_values(),
            dims=['mode'],
            coords=dict(mode=self._model._idx_mode[:n_rot]),
            name='singular_values'
        )

    def explained_covariance(self) -> DataArray:
        n_rot = self._params['n_rot']
        return xr.DataArray(
            super().explained_covariance(),
            dims=['mode'],
            coords=dict(mode=self._model._idx_mode[:n_rot]),
            name='explained_covariance'
        )

    def squared_covariance_fraction(self) -> DataArray:
        n_rot = self._params['n_rot']
        return xr.DataArray(
            super().squared_covariance_fraction(),
            dims=['mode'],
            coords=dict(mode=self._model._idx_mode[:n_rot]),
            name='squared_covariance_fraction'
        )

    def singular_vectors(self, scaling : int = 0) -> DataArrayList:
        Vx, Vy = super().singular_vectors(scaling=scaling)
        Vx = self._model._tfx.back_transform_eofs(Vx)
        Vy = self._model._tfy.back_transform_eofs(Vy)
        for x in Vx:
            x.name = 'left_singular_vectors'
        for y in Vy:
            y.name = 'right_singular_vectors'
        return squeeze(Vx), squeeze(Vy)

    def pcs(self, scaling : int = 0) -> DataArray:
        Ux, Uy = super().pcs(scaling=scaling)
        Ux = self._model._tfx.back_transform_pcs(Ux)
        Uy = self._model._tfy.back_transform_pcs(Uy)
        Ux.name = 'left_pcs'
        Uy.name = 'right_pcs'
        return Ux, Uy

    def homogeneous_patterns(self) -> Tuple[DataArrayList, DataArrayList]:
        hom_pats, pvals = super().homogeneous_patterns()
        hom_pats_x = self._model._tfx.back_transform_eofs(hom_pats[0])
        hom_pats_y = self._model._tfy.back_transform_eofs(hom_pats[1])
        pvals_x = self._model._tfx.back_transform_eofs(pvals[0])
        pvals_y = self._model._tfy.back_transform_eofs(pvals[1])

        for x, p in zip(hom_pats_x, pvals_x):
            x.name = 'left_homogeneous_patterns'
            p.name = 'left_homogeneous_patterns_p_values'
        for x, p in zip(hom_pats_y, pvals_y):
            x.name = 'right_homogeneous_patterns'
            p.name = 'right_homogeneous_patterns_p_values'

        hom_pats_x = squeeze(hom_pats_x)
        hom_pats_y = squeeze(hom_pats_y)
        pvals_x = squeeze(pvals_x)
        pvals_y = squeeze(pvals_y)
        return (hom_pats_x, hom_pats_y), (pvals_x, pvals_y)

    def heterogeneous_patterns(self) -> Tuple[DataArrayList, DataArrayList]:
        het_pats, pvals = super().heterogeneous_patterns()
        het_pats_x = self._model._tfx.back_transform_eofs(het_pats[0])
        het_pats_y = self._model._tfy.back_transform_eofs(het_pats[1])
        pvals_x = self._model._tfx.back_transform_eofs(pvals[0])
        pvals_y = self._model._tfy.back_transform_eofs(pvals[1])

        for x, p in zip(het_pats_x, pvals_x):
            x.name = 'left_heterogeneous_patterns'
            p.name = 'left_heterogeneous_patterns_p_values'
        for x, p in zip(het_pats_y, pvals_y):
            x.name = 'right_heterogeneous_patterns'
            p.name = 'right_heterogeneous_patterns_p_values'

        het_pats_x = squeeze(het_pats_x)
        het_pats_y = squeeze(het_pats_y)
        pvals_x = squeeze(pvals_x)
        pvals_y = squeeze(pvals_y)
        return (het_pats_x, het_pats_y), (pvals_x, pvals_y)

    def reconstruct_XY(
        self,
        mode : Optional[Union[int, List[int], slice]] = None
    ) -> DataArrayList:
        Xrec, Yrec = super().reconstruct_XY(mode=mode)
        Xrec = self._model._tfx.back_transform(Xrec)
        Yrec = self._model._tfy.back_transform(Yrec)
        return squeeze(Xrec), squeeze(Yrec)

    def project_onto_left_singular_vectors(
        self,
        X : DataArrayList,
        scaling : int = 0
    ) -> DataArray:
        proj = _MultiDataArrayTransformer()
        X = proj.fit_transform(X, dim=self._model._tfx.dims_samples)
        pcs = super().project_onto_left_singular_vectors(X=X, scaling=scaling)
        return proj.back_transform_pcs(pcs)

    def project_onto_right_singular_vectors(
        self,
        Y : DataArrayList,
        scaling : int = 0
    ) -> DataArray:
        proj = _MultiDataArrayTransformer()
        Y = proj.fit_transform(Y, dim=self._model._tfy.dims_samples)
        pcs = super().project_onto_right_singular_vectors(Y=Y, scaling=scaling)
        return proj.back_transform_pcs(pcs)



from typing import Optional, Tuple, List, Union

import numpy as np
import scipy as sc
from sklearn.decomposition import PCA

from ..utils.tools import get_mode_selector


class _BaseEOF():
    '''Base class for univariate EOF analysis.

    Parameters
    ----------
    X : np.ndarray
        2D data matrix ``X`` with samples as rows and features as columns.
        NaN entries will raise an error.
    n_modes : Optional[int]
        Number of modes to be retrieved. If None, then all possible modes will
        be computed. Reducing ``n_modes`` can greatly speed up computational
        (the default is None).
    norm : bool
        Normalize each feature by its standard deviation (the default is False).
    weights : Optional[np.ndarray]
        Apply `weights` to features. Weights must be a 1D array with the same
        length as number of features. No NaN entries are allowed
        (the default is None).

    Attributes
    ----------
    n_samples : int
        Number of samples (rows of data matrix).
    n_features : int
        Number of features (columns of data matrix).
    n_modes : int
        Number of modes.

    '''

    def __init__(
        self,
        X: np.ndarray,
        n_modes: Optional[int] = None,
        norm: bool = False,
        weights: Optional[np.ndarray] = None
    ):
        # Remove mean for each feature
        self._X_mean = X.mean(axis=0)
        X -= self._X_mean

        # Weights are applied to features, not samples.
        self._weights = weights
        if self._weights is None:
            # Use int type to ensure that there won't be rounding errors
            # when applying trivial weighting (= all weights equal 1)
            self._weights = np.ones(X.shape[1], dtype=int)

        # Standardization is included as weights
        if norm:
            stdev = X.std(axis=0)
            if (stdev == 0).any():
                err_msg = (
                    'Standard deviation of one ore more features is zero, '
                    'normalization not possible.'
                )
                raise ValueError(err_msg)
            self._weights = self._weights / stdev
        X = X * self._weights

        self.n_samples = X.shape[0]
        self.n_features = X.shape[1]

        self.n_modes = n_modes
        if n_modes is None:
            self.n_modes = min(self.n_samples, self.n_features)

        self.X = X

    def solve(self) -> None:
        '''
        Perform the EOF analysis.

        To boost performance, the standard solver is based on
        the PCA implementation of scikit-learn [1]_ which uses different algorithms
        to perform the decomposition based on the data matrix size.

        Naive approaches using singular value decomposition of the
        data matrix ``X (n x p)`` or the covariance matrix ``C (p x p)``
        quickly become infeasable computationally when the number of
        samples :math:`n` or features :math:`p` increase (computational power increases
        by :math:`O(n^2p)` and :math:`O(p^3)`, respectively.)


        References
        ----------
        .. [1] https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html

        '''

        pca = PCA(n_components=self.n_modes)
        self._pcs = pca.fit_transform(self.X)
        self._singular_values = pca.singular_values_
        self._explained_variance = pca.explained_variance_
        self._explained_variance_ratio  = pca.explained_variance_ratio_
        self._eofs = pca.components_.T

        # Normalize PCs so they are orthonormal
        # Note: singular values = sqrt(explained_variance * (n_samples - 1))
        self._pcs = self._pcs / self._singular_values

        # Ensure consistent signs for deterministic output
        maxidx = [abs(self._eofs).argmax(axis=0)]
        flip_signs = np.sign(self._eofs[maxidx, range(self._eofs.shape[1])])
        self._eofs *= flip_signs
        self._pcs *= flip_signs

    def singular_values(self) -> np.ndarray:
        '''Get the singular values.

        The `i` th singular value :math:`\sigma_i` is defined by

        .. math::
           \sigma_i = \sqrt{n \lambda_i}

        where :math:`\lambda_i` and :math:`n` are the associated eigenvalues
        and the number of samples, respectively.

        '''
        return self._singular_values

    def explained_variance(self) -> np.ndarray:
        '''Get the explained variance.

        The explained variance is simply given by the individual eigenvalues
        of the covariance matrix.

        '''
        return self._explained_variance

    def explained_variance_ratio(self) -> np.ndarray:
        '''Get the explained variance ratio.

        The explained variance ratio is the fraction of total variance
        explained by a given mode and is calculated by :math:`\lambda_i / \sum_i^m \lambda_i`
        where `m` is the total number of modes.

        '''

        return self._explained_variance_ratio

    def eofs(self, scaling : int = 0) -> np.ndarray:
        '''Get the EOFs.

        The empirical orthogonal functions (EOFs) are equivalent to the
        eigenvectors of the covariance matrix of `X`.

        Parameters
        ----------
        scaling : [0, 1, 2]
            EOFs are scaled (i) to be orthonormal (``scaling=0``), (ii) by the
            square root of the eigenvalues (``scaling=1``) or (iii) by the
            singular values (``scaling=2``). In case no weights were applied,
            scaling by the singular values results in the EOFs having the
            unit of the input data (the default is 0).

        '''
        if scaling == 0:
            eofs = self._eofs
        elif scaling == 1:
            eofs = self._eofs * np.sqrt(self._explained_variance)
        elif scaling == 2:
            eofs = self._eofs * self._singular_values
        return eofs

    def pcs(self, scaling : int = 0) -> np.ndarray:
        '''Get the PCs.

        The principal components (PCs), also known as PC scores, are computed
        by projecting the data matrix `X` onto the eigenvectors.

        Parameters
        ----------
        scaling : [0, 1, 2]
            PCs are scaled (i) to be orthonormal (``scaling=0``), (ii) by the
            square root of the eigenvalues (``scaling=1``) or (iii) by the
            singular values (``scaling=2``). In case no weights were applied,
            scaling by the singular values results in the PCs having the
            unit of the input data (the default is 0).

        '''
        if scaling == 0:
            pcs = self._pcs
        elif scaling == 1:
            pcs = self._pcs * np.sqrt(self._explained_variance)
        elif scaling == 2:
            pcs = self._pcs * self._singular_values
        return pcs

    def eofs_as_correlation(self) -> Tuple[np.ndarray, np.ndarray]:
        '''Correlation coefficients between PCs and data matrix.

        Returns
        -------
        Tuple[np.ndarray, np.ndarry]
            Matrices of correlation coefficients and associated
            two-sided p-values with features as rows and modes as columns.

        '''

        # Compute correlation matrix
        corr = np.corrcoef(self.X, self._pcs, rowvar=False)
        corr = corr[:self.n_features, self.n_features:]
        # Compute two-sided p-values
        # Reference: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.pearsonr.html#r8c6348c62346-1
        a = self.n_samples / 2 - 1
        dist = sc.stats.beta(a, a, loc=-1, scale=2)
        pvals = 2 * dist.cdf(-abs(corr))
        return corr, pvals

    def reconstruct_X(
        self,
        mode : Optional[Union[int, List[int], slice]] = None
    ) -> np.ndarray:
        '''Reconstruct original data field ``X`` using the PCs and EOFs.

        If weights were applied, ``X`` will be automatically rescaled.

        Parameters
        ----------
        mode : Optional[Union[int, List[int], slice]]
            Mode(s) based on which ``X`` will be reconstructed. If ``mode`` is
            an int, a single mode is used. If a list of integers is provided,
            use all specified modes for reconstruction. Alternatively, you may
            want to select a slice to reconstruct. The first mode is denoted
            by 1 (and not by 0). If None then ``X`` is recontructed using all
            available modes (the default is None).

        Examples
        --------

        Perform an analysis using some data ``X``:

        >>> model = EOF(X, norm=True)
        >>> model.solve()

        Reconstruct ``X`` using all modes:

        >>> model.reconstruct_X()

        Reconstruct ``X`` using the first mode only:

        >>> model.reconstruct_X(1)

        Reconstruct ``X`` using mode 1, 3 and 4:

        >>> model.reconstruct_X([1, 3, 4])

        Reconstruct ``X`` using all modes up to mode 10 (including):

        >>> model.reconstruct_X(slice(10))

        Reconstruct ``X`` using every second mode between 4 and 8 (both
        including):

        >>> model.reconstruct_X(slice(4, 8, 2))


        '''
        eofs = self._eofs
        pcs = self._pcs * self._singular_values
        # Select modes to reconstruct X
        mode = get_mode_selector(mode)
        eofs = eofs[:, mode]
        pcs = pcs[:, mode]
        Xrec = pcs @ eofs.T
        # Unweight and add mean
        return (Xrec / self._weights) + self._X_mean

    def project_onto_eofs(
        self,
        X : np.ndarray, scaling : int = 0
    ) -> np.ndarray:
        '''Project new data onto the EOFs.

        Parameters
        ----------
        X : np.ndarray
             New data to project. Data must be a 2D matrix.
        scaling : [0, 1, 2]
            Projections are scaled (i) to be orthonormal (``scaling=0``), (ii) by the
            square root of the eigenvalues (``scaling=1``) or (iii) by the
            singular values (``scaling=2``). In case no weights were applied,
            scaling by the singular values results in the projections having the
            unit of the input data (the default is 0).

        '''
        dof = self.n_samples - 1
        try:
            X -= self._X_mean
        except ValueError:
            err_msg = (
                'New data has invalid feature dimensions and cannot be '
                'projected onto EOFs.'
            )
            raise ValueError(err_msg)
        X *= self._weights
        pcs = X @ self._eofs / np.sqrt(self._explained_variance * dof)
        if scaling == 0:
            return pcs
        elif scaling == 1:
            return pcs * np.sqrt(self._explained_variance)
        elif scaling == 2:
            return pcs * np.sqrt(self._explained_variance * dof)
        else:
            err_msg = (
                'Scaling option {:} is not valid but must be one '
                'of [0, 1, 2]'
            )
            err_msg = err_msg.foramt(scaling)
            raise ValueError(err_msg)



from typing import Optional, Tuple, List, Union, Iterable

import numpy as np

from ._base_rock_pca import _BaseROCK_PCA
from xeofs.models._transformer import _MultiArrayTransformer
from ..utils.tools import squeeze

Array = np.ndarray
ArrayList = Union[Array, List[Array]]


class ROCK_PCA(_BaseROCK_PCA):
    '''ROCK-PCA of a single or multiple ``np.ndarray``.'''

    def __init__(
        self,
        X: np.ndarray,
        n_rot : int,
        power : int,
        sigma : float,
        axis : Union[int, Iterable[int]] = 0,
        n_modes: Optional[int] = None,
        norm: bool = False,
        weights: Optional[np.ndarray] = None,
    ):
        self._tf = _MultiArrayTransformer()
        X = self._tf.fit_transform(X, axis=axis)
        weights = self._tf.transform_weights(weights)

        super().__init__(
            X=X, n_rot=n_rot, power=power, sigma=sigma, n_modes=n_modes,
            norm=norm, weights=weights
        )

    def explained_variance(self) -> Array:
        return super().explained_variance()

    def explained_variance_ratio(self) -> Array:
        return super().explained_variance_ratio()

    def eofs(self) -> ArrayList:
        eofs = super().eofs()
        eofs = self._tf.back_transform_eofs(eofs)
        return squeeze(eofs)

    def pcs(self) -> Array:
        pcs = super().pcs()
        return self._tf.back_transform_pcs(pcs)

    def eofs_amplitude(self) -> ArrayList:
        amp = super().eofs_amplitude()
        amp = self._tf.back_transform_eofs(amp)
        return squeeze(amp)

    def pcs_amplitude(self) -> Array:
        amp = super().pcs_amplitude()
        return self._tf.back_transform_pcs(amp)

    def eofs_phase(self) -> ArrayList:
        phase = super().eofs_phase()
        phase = self._tf.back_transform_eofs(phase)
        return squeeze(phase)

    def pcs_phase(self) -> Array:
        phase = super().pcs_phase()
        return self._tf.back_transform_pcs(phase)
